#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import List, Union

from datasets import DatasetDict

from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.base import ArgillaTrainerSkeleton
from argilla.training.utils import _apply_column_mapping, filter_allowed_args, get_default_args
from argilla.utils.dependency import require_version


class ArgillaTransformersTrainer(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaTransformersTrainer")
    _logger.setLevel(logging.INFO)

    def __init__(self, *args, **kwargs):
        require_version("torch")
        require_version("datasets")
        require_version("transformers")
        require_version("evaluate")
        require_version("seqeval")
        super().__init__(*args, **kwargs)

        import torch
        from transformers import (
            AutoModelForSequenceClassification,
            AutoModelForTokenClassification,
            set_seed,
        )

        self._transformers_model = None
        self._transformers_tokenizer = None
        self._pipeline = None

        self.device = "cpu"
        if torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda"

        if self._seed is None:
            self._seed = 42
        set_seed(self._seed)

        if self._model is None:
            self._model = "bert-base-cased"

        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None

        if self._record_class == TextClassificationRecord:
            if self._multi_label:
                self._id2label = dict(enumerate(self._train_dataset.features["label"][0].names))
                self._label_list = self._train_dataset.features["label"][0].names
            else:
                self._id2label = dict(enumerate(self._train_dataset.features["label"].names))
                self._label_list = self._train_dataset.features["label"].names
            self._label2id = {v: k for k, v in self._id2label.items()}

            self._model_class = AutoModelForSequenceClassification

        elif self._record_class == TokenClassificationRecord:
            self._label_list = self._train_dataset.features["ner_tags"].feature.names
            self._id2label = dict(enumerate(self._label_list))
            self._label2id = {v: k for k, v in self._id2label.items()}

            self._model_class = AutoModelForTokenClassification
        else:
            raise NotImplementedError("Text2TextRecord is not supported.")

        self.init_training_args()

    def init_training_args(self):
        import torch
        from transformers import TrainingArguments

        def convert_to_floats(dataset):
            dataset.set_format("torch")
            return dataset.map(
                lambda x: {"float_label": x["label"].to(torch.float)}, remove_columns=["label"]
            ).rename_column("float_label", "label")

        if self._record_class == TextClassificationRecord:
            columns_mapping = {"text": "text", "label": "binarized_label"}
            if self._multi_label:
                self._train_dataset = _apply_column_mapping(self._train_dataset, columns_mapping)
                self._train_dataset = convert_to_floats(self._train_dataset)

                if self._eval_dataset is not None:
                    self._eval_dataset = _apply_column_mapping(self._eval_dataset, columns_mapping)
                    self._eval_dataset = convert_to_floats(self._eval_dataset)

        self.model_kwargs = {}
        self.model_kwargs["pretrained_model_name_or_path"] = self._model
        self.model_kwargs["num_labels"] = len(self._label_list)
        self.model_kwargs["id2label"] = self._id2label
        self.model_kwargs["label2id"] = self._label2id
        if self._multi_label:
            self.model_kwargs["problem_type"] = "multi_label_classification"

        self.trainer_kwargs = get_default_args(TrainingArguments.__init__)
        self.trainer_kwargs["evaluation_strategy"] = "no" if self._eval_dataset is None else "epoch"
        self.trainer_kwargs["logging_steps"] = 30
        self.trainer_kwargs["logging_steps"] = 1
        self.trainer_kwargs["num_train_epochs"] = 1

    def init_model(self, new: bool = False):
        from transformers import AutoTokenizer

        if any(k in self.model_kwargs.get("pretrained_model_name_or_path") for k in ("gpt", "opt", "bloom")):
            padding_side = "left"
        else:
            padding_side = "right"
        self._transformers_tokenizer = AutoTokenizer.from_pretrained(
            self.model_kwargs.get("pretrained_model_name_or_path"), padding_side=padding_side, add_prefix_space=True
        )
        if getattr(self._transformers_tokenizer, "pad_token_id") is None:
            self._transformers_tokenizer.pad_token_id = self._transformers_tokenizer.eos_token_id
        if getattr(self._transformers_tokenizer, "model_max_length") is None:
            self._transformers_tokenizer.model_max_length = 512

        if new:
            model_kwargs = self.model_kwargs
        else:
            model_kwargs = {"pretrained_model_name_or_path": self.model_kwargs.get("pretrained_model_name_or_path")}

        self._transformers_model = self._model_class.from_pretrained(**model_kwargs, return_dict=True)
        if new:
            self._transformers_model = self._transformers_model.to(self.device)

    def init_pipeline(self):
        import transformers
        from transformers import pipeline

        if self.device == "cuda":
            device = 0
        else:
            device = -1

        if self._record_class == TextClassificationRecord:
            if transformers.__version__ >= "4.20.0":
                kwargs = {"top_k": None}
            else:
                kwargs = {"return_all_scores": True}
            self._pipeline = pipeline(
                task="text-classification",
                model=self._transformers_model,
                tokenizer=self._transformers_tokenizer,
                device=device,
                **kwargs,
            )
        elif self._record_class == TokenClassificationRecord:
            self._pipeline = pipeline(
                task="token-classification",
                model=self._transformers_model,
                tokenizer=self._transformers_tokenizer,
                aggregation_strategy="first",
                device=device,
            )
        else:
            raise NotImplementedError("This is not implemented.")

    def update_config(self, **kwargs):
        """
        Updates the `setfit_model_kwargs` and `setfit_trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        from transformers import TrainingArguments

        self.trainer_kwargs.update(filter_allowed_args(TrainingArguments.__init__, **kwargs))

    def __repr__(self):
        formatted_string = []
        arg_dict = {
            "'AutoModel'": self.model_kwargs,
            "'Trainer'": self.trainer_kwargs,
        }
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\n".join(formatted_string)

    def preprocess_datasets(self):
        from transformers import (
            DataCollatorForTokenClassification,
            DataCollatorWithPadding,
        )

        def text_classification_preprocess_function(examples):
            return self._transformers_tokenizer(examples["text"], truncation=True, max_length=None)

        def token_classification_preprocess_function(examples):
            tokenized_inputs = self._transformers_tokenizer(
                examples["tokens"], truncation=True, is_split_into_words=True
            )

            labels = []
            for i, label in enumerate(examples["ner_tags"]):
                word_ids = tokenized_inputs.word_ids(batch_index=i)
                previous_word_idx = None
                label_ids = []
                for word_idx in word_ids:
                    if word_idx is None:
                        label_ids.append(-100)
                    elif word_idx != previous_word_idx:
                        label_ids.append(label[word_idx])
                    else:
                        label_ids.append(-100)
                    previous_word_idx = word_idx
                labels.append(label_ids)

            tokenized_inputs["labels"] = labels
            return tokenized_inputs

        # set correct tokenization
        if self._record_class == TextClassificationRecord:
            preprocess_function = text_classification_preprocess_function
            self._data_collator = DataCollatorWithPadding(tokenizer=self._transformers_tokenizer)
            if self._multi_label:
                remove_columns = ["feat_id", "text", "feat_label"]
            else:
                remove_columns = ["id", "text"]
            replace_labels = True
        elif self._record_class == TokenClassificationRecord:
            preprocess_function = token_classification_preprocess_function
            self._data_collator = DataCollatorForTokenClassification(tokenizer=self._transformers_tokenizer)
            remove_columns = ["id", "tokens", "ner_tags"]
            replace_labels = False
        else:
            raise NotImplementedError("`Text2TextRecord` is not supported yet.")

        self._tokenized_train_dataset = self._train_dataset.map(
            preprocess_function, batched=True, remove_columns=remove_columns
        )
        if replace_labels:
            self._tokenized_train_dataset = self._tokenized_train_dataset.rename_column("label", "labels")

        if self._eval_dataset is not None:
            self._tokenized_eval_dataset = self._eval_dataset.map(
                preprocess_function, batched=True, remove_columns=remove_columns
            )
            if replace_labels:
                self._tokenized_eval_dataset = self._tokenized_eval_dataset.rename_column("label", "labels")
        else:
            self._tokenized_eval_dataset = None

    def compute_metrics(self):
        import evaluate
        import numpy as np

        func = None
        if self._record_class == TextClassificationRecord:
            accuracy = evaluate.load("accuracy")
            f1 = evaluate.load("f1", config_name="multilabel")

            def compute_metrics_text_classification_multi_label(eval_pred):
                logits, labels = eval_pred
                # apply sigmoid
                predictions = (1.0 / (1 + np.exp(-logits))) > 0.5

                # f1 micro averaged
                metrics = f1.compute(predictions=predictions, references=labels, average="micro")
                # f1 per label
                per_label_metric = f1.compute(predictions=predictions, references=labels, average=None)
                for label, f1_score in zip(self._label_list, per_label_metric["f1"]):
                    metrics[f"f1_{label}"] = f1_score

                return metrics

            def compute_metrics_text_classification(eval_pred):
                predictions, labels = eval_pred
                predictions = np.argmax(predictions, axis=1)
                return accuracy.compute(predictions=predictions, references=labels)

            if self._multi_label:
                func = compute_metrics_text_classification_multi_label
            else:
                func = compute_metrics_text_classification

        elif self._record_class == TokenClassificationRecord:
            seqeval = evaluate.load("seqeval")

            def compute_metrics(p):
                predictions, labels = p
                predictions = np.argmax(predictions, axis=2)

                true_predictions = [
                    [self._label_list[p] for (p, l) in zip(prediction, label) if l != -100]
                    for prediction, label in zip(predictions, labels)
                ]
                true_labels = [
                    [self._label_list[l] for (p, l) in zip(prediction, label) if l != -100]
                    for prediction, label in zip(predictions, labels)
                ]

                results = seqeval.compute(predictions=true_predictions, references=true_labels)
                return {
                    "precision": results["overall_precision"],
                    "recall": results["overall_recall"],
                    "f1": results["overall_f1"],
                    "accuracy": results["overall_accuracy"],
                }

            func = compute_metrics
        else:
            raise NotImplementedError("Text2Text is not implemented.")
        return func

    def train(self, output_dir: str):
        from transformers import Trainer, TrainingArguments

        # check required path argument
        self.trainer_kwargs["output_dir"] = output_dir

        # prepare data
        self.init_model(new=True)
        self.preprocess_datasets()

        # get metrics function
        compute_metrics = self.compute_metrics()

        self._transformers_model.to(self.device)
        # set trainer
        self._training_args = TrainingArguments(**self.trainer_kwargs)
        self._transformers_trainer = Trainer(
            args=self._training_args,
            model=self._transformers_model,
            train_dataset=self._tokenized_train_dataset,
            eval_dataset=self._tokenized_eval_dataset,
            compute_metrics=compute_metrics,
            data_collator=self._data_collator,
        )

        #  train
        self._transformers_trainer.train()
        if self._tokenized_eval_dataset:
            self._metrics = self._transformers_trainer.evaluate()
            self._logger.info(self._metrics)
        else:
            self._metrics = None

        self.save(output_dir)

        self.init_pipeline()

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        if self._pipeline is None:
            self._logger.warning("Using model without fine-tuning.")
            self.init_model(new=False)
            self.init_pipeline()

        str_input = False
        if isinstance(text, str):
            text = [text]
            str_input = True

        predictions = self._pipeline(text, **kwargs)

        if as_argilla_records:
            formatted_prediction = []

            for val, pred in zip(text, predictions):
                if self._record_class == TextClassificationRecord:
                    formatted_prediction.append(
                        self._record_class(
                            text=val,
                            prediction=[(entry["label"], entry["score"]) for entry in pred],
                            multi_label=self._multi_label,
                        )
                    )
                elif self._record_class == TokenClassificationRecord:
                    _pred = [(value["entity_group"], value["start"], value["end"]) for value in pred]
                    encoding = self._pipeline.tokenizer(val)
                    word_ids = sorted(set(encoding.word_ids()) - {None})
                    tokens = []
                    for word_id in word_ids:
                        char_span = encoding.word_to_chars(word_id)
                        tokens.append(val[char_span.start : char_span.end].lstrip().rstrip())
                    formatted_prediction.append(
                        self._record_class(
                            text=val,
                            tokens=tokens,
                            prediction=_pred,
                        )
                    )

                else:
                    raise NotImplementedError("This is not implemented yet.")
        else:
            formatted_prediction = predictions

        if str_input:
            formatted_prediction = formatted_prediction[0]

        return formatted_prediction

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified, and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self._transformers_model.save_pretrained(output_dir)
        self._transformers_tokenizer.save_pretrained(output_dir)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""create users table

Revision ID: 74694870197c
Revises:
Create Date: 2023-02-13 17:08:05.445314

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "74694870197c"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "users",
        sa.Column("id", sa.Uuid, primary_key=True),
        sa.Column("first_name", sa.String, nullable=False),
        sa.Column("last_name", sa.String),
        sa.Column("username", sa.String, nullable=False, unique=True, index=True),
        sa.Column("role", sa.String, nullable=False, index=True),
        sa.Column("api_key", sa.Text, nullable=False, unique=True, index=True),
        sa.Column("password_hash", sa.Text, nullable=False),
        sa.Column("inserted_at", sa.DateTime, nullable=False),
        sa.Column("updated_at", sa.DateTime, nullable=False),
    )


def downgrade() -> None:
    op.drop_table("users")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query, Security, status
from fastapi import Response as HTTPResponse
from sqlalchemy.ext.asyncio import AsyncSession

from argilla.server.contexts import datasets
from argilla.server.database import get_async_db
from argilla.server.models import User
from argilla.server.policies import RecordPolicyV1, authorize
from argilla.server.schemas.v1.datasets import Record as RecordSchema
from argilla.server.schemas.v1.records import Response, ResponseCreate
from argilla.server.schemas.v1.suggestions import Suggestion, SuggestionCreate, Suggestions
from argilla.server.search_engine import SearchEngine, get_search_engine
from argilla.server.security import auth
from argilla.server.utils import parse_uuids

if TYPE_CHECKING:
    from argilla.server.models import Record

DELETE_RECORD_SUGGESTIONS_LIMIT = 100

router = APIRouter(tags=["records"])


async def _get_record(
    db: AsyncSession, record_id: UUID, with_dataset: bool = False, with_suggestions: bool = False
) -> "Record":
    record = await datasets.get_record_by_id(db, record_id, with_dataset, with_suggestions)
    if not record:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Record with id `{record_id}` not found",
        )
    return record


@router.post("/records/{record_id}/responses", status_code=status.HTTP_201_CREATED, response_model=Response)
async def create_record_response(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    response_create: ResponseCreate,
    current_user: User = Security(auth.get_current_user),
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.create_response(record))

    if await datasets.get_response_by_record_id_and_user_id(db, record_id, current_user.id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Response already exists for record with id `{record_id}` and by user with id `{current_user.id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return await datasets.create_response(db, search_engine, record, current_user, response_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.get("/records/{record_id}/suggestions", status_code=status.HTTP_200_OK, response_model=Suggestions)
async def get_record_suggestions(
    *, db: AsyncSession = Depends(get_async_db), record_id: UUID, current_user: User = Security(auth.get_current_user)
):
    record = await _get_record(db, record_id, with_dataset=True, with_suggestions=True)

    await authorize(current_user, RecordPolicyV1.get_suggestions(record))

    return Suggestions(items=record.suggestions)


@router.put(
    "/records/{record_id}/suggestions",
    summary="Create or update a suggestion",
    responses={
        status.HTTP_200_OK: {"model": Suggestion, "description": "Suggestion updated"},
        status.HTTP_201_CREATED: {"model": Suggestion, "description": "Suggestion created"},
    },
    status_code=status.HTTP_201_CREATED,
    response_model=Suggestion,
)
async def upsert_suggestion(
    *,
    db: AsyncSession = Depends(get_async_db),
    record_id: UUID,
    suggestion_create: SuggestionCreate,
    current_user: User = Security(auth.get_current_user),
    response: HTTPResponse,
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.create_suggestion(record))

    question = await datasets.get_question_by_id(db, suggestion_create.question_id)
    if not question:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Question with id `{suggestion_create.question_id}` not found",
        )

    if await datasets.get_suggestion_by_record_id_and_question_id(db, record_id, suggestion_create.question_id):
        # There is already a suggestion for this record and question, so we update it.
        response.status_code = status.HTTP_200_OK

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return await datasets.upsert_suggestion(db, record, question, suggestion_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.delete(
    "/records/{record_id}/suggestions",
    summary="Delete suggestions for a record",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_record_suggestions(
    *,
    db: AsyncSession = Depends(get_async_db),
    record_id: UUID,
    current_user: User = Security(auth.get_current_user),
    ids: str = Query(..., description="A comma separated list with the IDs of the suggestions to be removed"),
):
    record = await _get_record(db, record_id)

    await authorize(current_user, RecordPolicyV1.delete_suggestions(record))

    suggestion_ids = parse_uuids(ids)
    num_suggestions = len(suggestion_ids)

    if num_suggestions == 0:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="No suggestions IDs provided")

    if num_suggestions > DELETE_RECORD_SUGGESTIONS_LIMIT:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Cannot delete more than {DELETE_RECORD_SUGGESTIONS_LIMIT} suggestions at once",
        )

    await datasets.delete_suggestions(db, record, suggestion_ids)


@router.delete("/records/{record_id}", response_model=RecordSchema, response_model_exclude_unset=True)
async def delete_record(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.delete(record))

    return await datasets.delete_record(db, search_engine, record)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Iterable, List, Optional, Set, Type, Union

from fastapi import Depends

from argilla.server.daos.models.records import DaoRecordsSearch
from argilla.server.daos.records import DatasetRecordsDAO
from argilla.server.errors import RecordNotFound
from argilla.server.services.datasets import ServiceDataset
from argilla.server.services.metrics import MetricsService
from argilla.server.services.metrics.models import ServiceMetric
from argilla.server.services.search.model import (
    ServiceRecord,
    ServiceRecordsQuery,
    ServiceSearchResults,
    ServiceSortConfig,
)


class SearchRecordsService:
    """Generic service for search records operations"""

    _INSTANCE: "SearchRecordsService" = None

    __LOGGER__ = logging.getLogger(__name__)

    @classmethod
    def get_instance(
        cls,
        dao: DatasetRecordsDAO = Depends(DatasetRecordsDAO.get_instance),
        metrics: MetricsService = Depends(MetricsService.get_instance),
    ):
        if not cls._INSTANCE:
            cls._INSTANCE = cls(dao=dao, metrics=metrics)
        return cls._INSTANCE

    def __init__(
        self,
        dao: DatasetRecordsDAO,
        metrics: MetricsService,
    ):
        self.__dao__ = dao
        self.__metrics__ = metrics

    def search(
        self,
        dataset: ServiceDataset,
        record_type: Type[ServiceRecord],
        query: Optional[ServiceRecordsQuery] = None,
        sort_config: Optional[ServiceSortConfig] = None,
        record_from: int = 0,
        size: int = 100,
        exclude_metrics: bool = True,
        metrics: Optional[List[ServiceMetric]] = None,
    ) -> ServiceSearchResults:
        if record_from > 0:
            metrics = None

        sort_config = sort_config or ServiceSortConfig()
        exclude_fields = ["metrics.*"] if exclude_metrics else None
        if query and query.vector and not query.vector.k:
            query.vector.k = size
        results = self.__dao__.search_records(
            dataset,
            search=DaoRecordsSearch(
                query=query,
                sort=sort_config,
            ),
            size=size,
            record_from=record_from,
            exclude_fields=exclude_fields,
            highligth_results=query is not None and query.query_text is not None and len(query.query_text) > 0,
        )
        metrics_results = {}
        for metric in metrics or []:
            try:
                metrics_ = self.__metrics__.summarize_metric(
                    dataset=dataset,
                    metric=metric,
                    record_class=record_type,
                    query=query,
                )
                metrics_results[metric.id] = metrics_
            except Exception as ex:
                self.__LOGGER__.warning("Cannot compute metric [%s]. Error: %s", metric.id, ex)
                metrics_results[metric.id] = {}

        return ServiceSearchResults(
            total=results.total,
            records=[record_type.parse_obj(r) for r in results.records],
            metrics=metrics_results if metrics_results else {},
        )

    async def find_record_by_id(
        self,
        dataset: ServiceDataset,
        id: Union[str, int],
        record_type: Type[ServiceRecord],
    ) -> ServiceRecord:
        found = await self.__dao__.get_record_by_id(dataset, id)
        if not found:
            raise RecordNotFound(
                dataset=dataset.id,
                id=id,
                type="Record",
            )

        return record_type.parse_obj(found)

    def scan_records(
        self,
        dataset: ServiceDataset,
        record_type: Optional[Type[ServiceRecord]] = None,
        query: Optional[ServiceRecordsQuery] = None,
        projection: Set[str] = None,
        id_from: Optional[str] = None,
        limit: int = 1000,
    ) -> Iterable[ServiceRecord]:
        """Scan records for a queried"""
        search = DaoRecordsSearch(query=query)

        transform_doc = lambda doc: doc
        if record_type:
            transform_doc = record_type.parse_obj

        for doc in self.__dao__.scan_dataset(
            dataset,
            id_from=id_from,
            limit=limit,
            search=search,
            include_fields=projection,
        ):
            yield transform_doc(doc)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from argilla.server.commons.models import PredictionStatus, TaskType
from argilla.server.services.datasets import ServiceBaseDataset
from argilla.server.services.search.model import ServiceBaseRecordsQuery, ServiceScoreRange
from argilla.server.services.tasks.commons import ServiceBaseAnnotation, ServiceBaseRecord


class ServiceText2TextPrediction(BaseModel):
    text: str
    score: float


class ServiceText2TextAnnotation(ServiceBaseAnnotation):
    sentences: List[ServiceText2TextPrediction]


class ServiceText2TextRecord(ServiceBaseRecord[ServiceText2TextAnnotation]):
    text: str

    @classmethod
    def task(cls) -> TaskType:
        """The task type"""
        return TaskType.text2text

    def all_text(self) -> str:
        return self.text

    @property
    def predicted_as(self) -> Optional[List[str]]:
        return [sentence.text for sentence in self.prediction.sentences] if self.prediction else None

    @property
    def annotated_as(self) -> Optional[List[str]]:
        return [sentence.text for sentence in self.annotation.sentences] if self.annotation else None

    @property
    def scores(self) -> List[float]:
        """Values of prediction scores"""
        if not self.prediction:
            return []
        return [sentence.score for sentence in self.prediction.sentences]

    def extended_fields(self) -> Dict[str, Any]:
        return {
            "annotated_as": self.annotated_as,
            "predicted_as": self.predicted_as,
            "annotated_by": self.annotated_by,
            "predicted_by": self.predicted_by,
            "score": self.scores,
        }


class ServiceText2TextQuery(ServiceBaseRecordsQuery):
    score: Optional[ServiceScoreRange] = Field(default=None)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
from typing import List, Union

import numpy as np

from argilla._constants import OPENAI_END_TOKEN, OPENAI_SEPARATOR, OPENAI_WHITESPACE
from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.datasets import TextClassificationSettings, TokenClassificationSettings
from argilla.training.base import ArgillaTrainerSkeleton
from argilla.training.utils import filter_allowed_args
from argilla.utils.dependency import require_version


class ArgillaOpenAITrainer(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaOpenAITrainer")
    _logger.setLevel(logging.INFO)
    _separator = OPENAI_SEPARATOR
    _end_token = OPENAI_END_TOKEN
    _whitespace = OPENAI_WHITESPACE

    OPENAI_API_KEY = "OPENAI_API_KEY"
    if OPENAI_API_KEY not in os.environ:
        raise ValueError(f"{OPENAI_API_KEY} not found in environment variables.")

    def __init__(self, *args, **kwargs):
        require_version("openai")
        super().__init__(*args, **kwargs)

        if self._record_class is TokenClassificationRecord:
            raise NotImplementedError("OpenAI does not support `TokenClassification` tasks.")
        elif self._record_class is TextClassificationRecord and self._multi_label:
            raise NotImplementedError("OpenAI does not support `multi-label=True` TextClassification tasks.")
        else:
            pass

        self.sleep_timer = 10
        self.device = None
        self.finetune_id = None

        if self._seed is not None:
            self._logger.warning("Seed is not supported for OpenAI. Ignoring seed for training.")

        if self._model is None:
            self._model = "curie"

        if isinstance(self._dataset, tuple):
            self._train_dataset = self._dataset[0]
            self._eval_dataset = self._dataset[1]
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None

        self.init_training_args(model=self._model)

    def init_training_args(
        self,
        training_file: str = None,
        validation_file: str = None,
        model: str = "curie",
        n_epochs: int = None,
        batch_size: int = None,
        learning_rate_multiplier: float = 0.1,
        prompt_loss_weight: float = 0.1,
        compute_classification_metrics: bool = False,
        classification_n_classes: int = None,
        classification_positive_class: str = None,
        classification_betas: list = None,
        suffix: str = None,
    ):
        self.model_kwargs = {}

        self.model_kwargs["training_file"] = training_file
        self.model_kwargs["validation_file"] = validation_file
        self.model_kwargs["model"] = model
        if isinstance(self._record_class, TextClassificationRecord):
            self.model_kwargs["n_epochs"] = n_epochs or 4
        else:
            self.model_kwargs["n_epochs"] = n_epochs or 2
        self.model_kwargs["batch_size"] = batch_size
        self.model_kwargs["learning_rate_multiplier"] = learning_rate_multiplier
        self.model_kwargs["prompt_loss_weight"] = prompt_loss_weight
        self.model_kwargs["compute_classification_metrics"] = compute_classification_metrics
        self.model_kwargs["classification_n_classes"] = classification_n_classes
        self.model_kwargs["classification_positive_class"] = classification_positive_class
        self.model_kwargs["classification_betas"] = classification_betas
        self.model_kwargs["suffix"] = suffix

        if isinstance(self._record_class, TextClassificationRecord) and self._eval_dataset:
            label_schema = self._label_list
            if len(label_schema) == 2:
                self.model_kwargs["classification_positive_class"] = label_schema[0]
                self.model_kwargs["compute_classification_metrics"] = True
            else:
                self.model_kwargs["classification_n_classes"] = len(label_schema)
                self.model_kwargs["compute_classification_metrics"] = True

    def update_config(
        self,
        **kwargs,
    ):
        """
        Updates the `model_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        self.model_kwargs.update(kwargs)
        self.model_kwargs = filter_allowed_args(self.init_training_args, **self.model_kwargs)

        keys = []
        for key, value in self.model_kwargs.items():
            if value is None:
                keys.append(key)
        for key in keys:
            del self.model_kwargs[key]

        if "model" in self.model_kwargs:
            self._model = self.model_kwargs["model"]

    def __repr__(self):
        formatted_string = []
        arg_dict = {
            "'OpenAI.fine_tune'": self.model_kwargs,
        }
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\nhttps://platform.openai.com/docs/api-reference/fine-tune\n" + "\n".join(formatted_string)

    def upload_dataset_to_openai(self, dataset, file_name):
        import json

        import openai

        for rec in dataset:
            if "id" in rec:
                del rec["id"]
        with open(file_name, "w") as f:
            for item in dataset:
                f.write(json.dumps(item) + "\n")
        upload_response = openai.File.create(file=open(file_name, "rb"), purpose="fine-tune")
        self._logger.info(upload_response)
        file_id = upload_response.id
        return file_id

    def train(self, output_dir: str = None):
        """
        We create a openai.FineTune object from a pretrained model, and send data to finetune it.
        """

        import openai

        if output_dir is not None:
            self.model_kwargs["suffix"] = output_dir

        if self._train_dataset is not None and self.model_kwargs["training_file"] is None:
            self.model_kwargs["training_file"] = self.upload_dataset_to_openai(self._train_dataset, "data_train.jsonl")
        if (self._eval_dataset is not None and self.model_kwargs["validation_file"] is None) and self.model_kwargs[
            "compute_classification_metrics"
        ]:
            self.model_kwargs["validation_file"] = self.upload_dataset_to_openai(self._eval_dataset, "data_test.jsonl")

        self.update_config()

        response = openai.FineTune.create(
            **self.model_kwargs,
        )

        self._logger.info(response)
        self.finetune_id = response.id
        self._logger.info(f"Waiting for training. Get info by running: `openai.FineTune.retrieve({self.finetune_id})`.")

    def init_model(self):
        import openai

        if self.finetune_id is not None:
            response = openai.FineTune.retrieve(self.finetune_id)
            potential_model = response.fine_tuned_model
            if potential_model is None:
                self._logger.warning("Fine-tuning is still in progress.")
            else:
                self._model = potential_model
        else:
            response = openai.Model.retrieve(self._model)
            self._model = response.id
            self._logger.info(response)

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """

        import openai

        responses = []
        was_string = False

        if isinstance(text, str):
            text = [text]
            was_string = True

        if isinstance(self._settings, TextClassificationSettings):
            for kwarg in ["logprobs", "max_tokens", "temperature", "n"]:
                if kwarg in kwargs:
                    del kwargs[kwarg]
                    self._logger.warning(f"Argument `{kwarg}` has default value for text classification. Deleting it.")
            kwargs["logprobs"] = len(self._settings.label_schema)
            kwargs["max_tokens"] = 1
            kwargs["temperature"] = 0
            kwargs["n"] = 1
        else:
            if "stop" in kwargs:
                del kwargs[kwarg]
            self._logger.warning("Argument `stop` has default value for text classification. Deleting it.")
            kwargs["stop"] = self._end_token
            if "logprobs" not in kwargs:
                kwargs["logprobs"] = 1

        for entry in text:
            prompt = f"{entry.strip()}{self._separator}"
            response = openai.Completion.create(model=self._model, prompt=prompt, **kwargs)
            if isinstance(self._settings, TextClassificationSettings):
                logprobs = response["choices"][0]["logprobs"]["top_logprobs"][0]
                keys = [self._settings.id2label[int(key.strip())] for key in list(logprobs.keys())]
                values = np.exp(list(logprobs.values()))
                response["choices"][0]["logprobs"]["top_logprobs"][0] = dict(zip(keys, values))
                if as_argilla_records:
                    response = self._record_class(text=entry, prediction=list(zip(keys, values)))
            elif isinstance(self._settings, TokenClassificationSettings):
                raise NotImplementedError("TokenClassification is not supported yet.")
            else:
                if as_argilla_records:
                    predictions = [choice["text"] for choice in response["choices"]]
                    response = self._record_class(text=entry, prediction=predictions)

            responses.append(response)

        if was_string:
            return responses[0]
        else:
            return responses

    def save(self, *arg, **kwargs):
        """
        The function saves the model to the path specified, and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self._logger.warning("Saving is not supported for OpenAI and is passed via the `suffix` argument in `train`.")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import UUID

from fastapi.encoders import jsonable_encoder
from sqlalchemy import and_, func, or_, select
from sqlalchemy.orm import contains_eager, joinedload, selectinload

from argilla.server.contexts import accounts
from argilla.server.enums import RecordInclude, ResponseStatusFilter
from argilla.server.models import (
    Dataset,
    DatasetStatus,
    Field,
    Question,
    Record,
    Response,
    ResponseStatus,
    ResponseValue,
    Suggestion,
)
from argilla.server.models.suggestions import SuggestionCreateWithRecordId
from argilla.server.schemas.v1.datasets import DatasetCreate, FieldCreate, QuestionCreate, RecordsCreate
from argilla.server.schemas.v1.records import ResponseCreate
from argilla.server.schemas.v1.responses import ResponseUpdate
from argilla.server.search_engine import SearchEngine
from argilla.server.security.model import User

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession

    from argilla.server.schemas.v1.datasets import DatasetUpdate
    from argilla.server.schemas.v1.fields import FieldUpdate
    from argilla.server.schemas.v1.questions import QuestionUpdate
    from argilla.server.schemas.v1.suggestions import SuggestionCreate

LIST_RECORDS_LIMIT = 20


async def get_dataset_by_id(
    db: "AsyncSession", dataset_id: UUID, with_fields: bool = False, with_questions: bool = False
) -> Dataset:
    query = select(Dataset).filter_by(id=dataset_id)
    options = []
    if with_fields:
        options.append(selectinload(Dataset.fields))
    if with_questions:
        options.append(selectinload(Dataset.questions))
    if options:
        query = query.options(*options)
    result = await db.execute(query)
    return result.scalar_one_or_none()


async def get_dataset_by_name_and_workspace_id(
    db: "AsyncSession", name: str, workspace_id: UUID
) -> Union[Dataset, None]:
    result = await db.execute(select(Dataset).filter_by(name=name, workspace_id=workspace_id))
    return result.scalar_one_or_none()


async def list_datasets(db: "AsyncSession") -> List[Dataset]:
    result = await db.execute(select(Dataset).order_by(Dataset.inserted_at.asc()))
    return result.scalars().all()


async def list_datasets_by_workspace_id(db: "AsyncSession", workspace_id: UUID) -> List[Dataset]:
    result = await db.execute(
        select(Dataset).where(Dataset.workspace_id == workspace_id).order_by(Dataset.inserted_at.asc())
    )
    return result.scalars().all()


async def create_dataset(db: "AsyncSession", dataset_create: DatasetCreate):
    return await Dataset.create(
        db,
        name=dataset_create.name,
        guidelines=dataset_create.guidelines,
        workspace_id=dataset_create.workspace_id,
    )


async def _count_required_fields_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(Field.id)).filter_by(dataset_id=dataset_id, required=True))
    return result.scalar()


async def _count_required_questions_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(Question.id)).filter_by(dataset_id=dataset_id, required=True))
    return result.scalar()


async def publish_dataset(db: "AsyncSession", search_engine: SearchEngine, dataset: Dataset) -> Dataset:
    if dataset.is_ready:
        raise ValueError("Dataset is already published")

    if await _count_required_fields_by_dataset_id(db, dataset.id) == 0:
        raise ValueError("Dataset cannot be published without required fields")

    if await _count_required_questions_by_dataset_id(db, dataset.id) == 0:
        raise ValueError("Dataset cannot be published without required questions")

    async with db.begin_nested():
        dataset = await dataset.update(db, status=DatasetStatus.ready, autocommit=False)
        await search_engine.create_index(dataset)

    await db.commit()

    return dataset


async def delete_dataset(db: "AsyncSession", search_engine: SearchEngine, dataset: Dataset) -> Dataset:
    async with db.begin_nested():
        dataset = await dataset.delete(db, autocommit=False)
        await search_engine.delete_index(dataset)

    await db.commit()

    return dataset


async def update_dataset(db: "AsyncSession", dataset: Dataset, dataset_update: "DatasetUpdate") -> Dataset:
    params = dataset_update.dict(exclude_unset=True)
    return await dataset.update(db, **params)


async def get_field_by_id(db: "AsyncSession", field_id: UUID) -> Union[Field, None]:
    result = await db.execute(select(Field).filter_by(id=field_id).options(selectinload(Field.dataset)))
    return result.scalar_one_or_none()


async def get_field_by_name_and_dataset_id(db: "AsyncSession", name: str, dataset_id: UUID) -> Union[Field, None]:
    result = await db.execute(select(Field).filter_by(name=name, dataset_id=dataset_id))
    return result.scalar_one_or_none()


async def create_field(db: "AsyncSession", dataset: Dataset, field_create: FieldCreate) -> Field:
    if dataset.is_ready:
        raise ValueError("Field cannot be created for a published dataset")

    return await Field.create(
        db,
        name=field_create.name,
        title=field_create.title,
        required=field_create.required,
        settings=field_create.settings.dict(),
        dataset_id=dataset.id,
    )


async def update_field(db: "AsyncSession", field: Field, field_update: "FieldUpdate") -> Field:
    params = field_update.dict(exclude_unset=True)
    return await field.update(db, **params)


async def delete_field(db: "AsyncSession", field: Field) -> Field:
    if field.dataset.is_ready:
        raise ValueError("Fields cannot be deleted for a published dataset")

    return await field.delete(db)


async def get_question_by_id(db: "AsyncSession", question_id: UUID) -> Union[Question, None]:
    result = await db.execute(select(Question).filter_by(id=question_id).options(selectinload(Question.dataset)))
    return result.scalar_one_or_none()


async def get_question_by_name_and_dataset_id(db: "AsyncSession", name: str, dataset_id: UUID) -> Union[Question, None]:
    result = await db.execute(select(Question).filter_by(name=name, dataset_id=dataset_id))
    return result.scalar_one_or_none()


async def create_question(db: "AsyncSession", dataset: Dataset, question_create: QuestionCreate) -> Question:
    if dataset.is_ready:
        raise ValueError("Question cannot be created for a published dataset")

    return await Question.create(
        db,
        name=question_create.name,
        title=question_create.title,
        description=question_create.description,
        required=question_create.required,
        settings=question_create.settings.dict(),
        dataset_id=dataset.id,
    )


async def update_question(db: "AsyncSession", question: Question, question_update: "QuestionUpdate") -> Question:
    params = question_update.dict(exclude_unset=True)
    return await question.update(db, **params)


async def delete_question(db: "AsyncSession", question: Question) -> Question:
    if question.dataset.is_ready:
        raise ValueError("Questions cannot be deleted for a published dataset")

    return await question.delete(db)


async def get_record_by_id(
    db: "AsyncSession", record_id: UUID, with_dataset: bool = False, with_suggestions: bool = False
) -> Union[Record, None]:
    query = select(Record).filter_by(id=record_id)
    if with_dataset:
        query = query.options(selectinload(Record.dataset).selectinload(Dataset.questions))
    if with_suggestions:
        query = query.options(selectinload(Record.suggestions))
    result = await db.execute(query)
    return result.scalar_one_or_none()


async def delete_record(db: "AsyncSession", search_engine: "SearchEngine", record: Record) -> Record:
    async with db.begin_nested():
        record = await record.delete(db=db, autocommit=False)
        await search_engine.delete_records(dataset=record.dataset, records=[record])

    await db.commit()

    return record


async def delete_records(
    db: "AsyncSession", search_engine: "SearchEngine", dataset: Dataset, records_ids: List[UUID]
) -> None:
    async with db.begin_nested():
        params = [Record.id.in_(records_ids), Record.dataset_id == dataset.id]
        records = await Record.delete_many(db=db, params=params, autocommit=False)
        await search_engine.delete_records(dataset=dataset, records=records)

    await db.commit()


async def get_records_by_ids(
    db: "AsyncSession",
    dataset_id: UUID,
    record_ids: List[UUID],
    include: List[RecordInclude] = [],
    user_id: Optional[UUID] = None,
) -> List[Record]:
    query = select(Record).filter(Record.dataset_id == dataset_id, Record.id.in_(record_ids))
    if RecordInclude.responses in include:
        if user_id:
            query = query.outerjoin(
                Response, and_(Response.record_id == Record.id, Response.user_id == user_id)
            ).options(contains_eager(Record.responses))
        else:
            query = query.options(joinedload(Record.responses))
    if RecordInclude.suggestions in include:
        query = query.options(joinedload(Record.suggestions))
    result = await db.execute(query)
    return result.unique().scalars().all()


async def list_records_by_dataset_id(
    db: "AsyncSession",
    dataset_id: UUID,
    user_id: Optional[UUID] = None,
    include: List[RecordInclude] = [],
    response_statuses: List[ResponseStatusFilter] = [],
    offset: int = 0,
    limit: int = LIST_RECORDS_LIMIT,
) -> List[Record]:
    response_statuses_ = [
        ResponseStatus(response_status)
        for response_status in response_statuses
        if response_status != ResponseStatusFilter.missing
    ]

    response_status_filter_expressions = []

    if response_statuses_:
        response_status_filter_expressions.append(Response.status.in_(response_statuses_))

    if ResponseStatusFilter.missing in response_statuses:
        response_status_filter_expressions.append(Response.status.is_(None))

    query = (
        select(Record)
        .filter(Record.dataset_id == dataset_id)
        .outerjoin(
            Response,
            Response.record_id == Record.id
            if user_id is None
            else and_(Response.record_id == Record.id, Response.user_id == user_id),
        )
    )

    if response_status_filter_expressions:
        query = query.filter(or_(*response_status_filter_expressions))

    if RecordInclude.responses in include:
        query = query.options(contains_eager(Record.responses))

    if RecordInclude.suggestions in include:
        query = query.options(joinedload(Record.suggestions))

    query = query.order_by(Record.inserted_at.asc()).offset(offset).limit(limit)
    result = await db.execute(query)
    return result.unique().scalars().all()


async def count_records_by_dataset_id(db: "AsyncSession", dataset_id: UUID) -> int:
    result = await db.execute(select(func.count(Record.id)).filter_by(dataset_id=dataset_id))
    return result.scalar()


async def create_records(
    db: "AsyncSession", search_engine: SearchEngine, dataset: Dataset, records_create: RecordsCreate
):
    if not dataset.is_ready:
        raise ValueError("Records cannot be created for a non published dataset")

    records = []
    for record_create in records_create.items:
        validate_record_fields(dataset, fields=record_create.fields)

        record = Record(
            fields=record_create.fields,
            metadata_=record_create.metadata,
            external_id=record_create.external_id,
            dataset_id=dataset.id,
        )

        if record_create.responses:
            for response in record_create.responses:
                # TODO(gabrielmbmb): the result of this query should be cached
                if not await accounts.get_user_by_id(db, response.user_id):
                    raise ValueError(f"Provided user_id: {response.user_id!r} is not a valid user id")

                validate_response_values(dataset, values=response.values, status=response.status)

                record.responses.append(
                    Response(
                        values=jsonable_encoder(response.values),
                        status=response.status,
                        user_id=response.user_id,
                    )
                )

        if record_create.suggestions:
            for suggestion in record_create.suggestions:
                # TODO(gabrielmbmb): the result of this query should be cached
                question = await get_question_by_id(db, suggestion.question_id)
                if not question:
                    raise ValueError(f"Provided question_id: {suggestion.question_id!r} is not a valid question id")

                question.parsed_settings.check_response(suggestion)

                record.suggestions.append(
                    Suggestion(
                        type=suggestion.type,
                        score=suggestion.score,
                        value=suggestion.value,
                        agent=suggestion.agent,
                        question_id=suggestion.question_id,
                    )
                )

        records.append(record)

    async with db.begin_nested():
        db.add_all(records)
        await db.flush(records)
        for record in records:
            await record.awaitable_attrs.responses
        await search_engine.add_records(dataset, records)

    await db.commit()


async def get_response_by_id(db: "AsyncSession", response_id: UUID) -> Union[Response, None]:
    result = await db.execute(
        select(Response)
        .filter_by(id=response_id)
        .options(selectinload(Response.record).selectinload(Record.dataset).selectinload(Dataset.questions))
    )
    return result.scalar_one_or_none()


async def get_response_by_record_id_and_user_id(
    db: "AsyncSession", record_id: UUID, user_id: UUID
) -> Union[Response, None]:
    result = await db.execute(select(Response).filter_by(record_id=record_id, user_id=user_id))
    return result.scalar_one_or_none()


async def count_responses_by_dataset_id_and_user_id(
    db: "AsyncSession", dataset_id: UUID, user_id: UUID, response_status: Optional[ResponseStatus] = None
) -> int:
    expressions = [Response.user_id == user_id]
    if response_status:
        expressions.append(Response.status == response_status)

    result = await db.execute(
        select(func.count(Response.id))
        .join(Record, and_(Record.id == Response.record_id, Record.dataset_id == dataset_id))
        .filter(*expressions)
    )
    return result.scalar()


async def create_response(
    db: "AsyncSession", search_engine: SearchEngine, record: Record, user: User, response_create: ResponseCreate
) -> Response:
    validate_response_values(record.dataset, values=response_create.values, status=response_create.status)

    async with db.begin_nested():
        response = await Response.create(
            db,
            values=jsonable_encoder(response_create.values),
            status=response_create.status,
            record_id=record.id,
            user_id=user.id,
            autocommit=False,
        )
        await db.flush([response])
        await search_engine.update_record_response(response)

    await db.commit()

    return response


async def update_response(
    db: "AsyncSession", search_engine: SearchEngine, response: Response, response_update: ResponseUpdate
):
    validate_response_values(response.record.dataset, values=response_update.values, status=response_update.status)

    async with db.begin_nested():
        response = await response.update(
            db, values=jsonable_encoder(response_update.values), status=response_update.status, autocommit=False
        )
        await search_engine.update_record_response(response)

    await db.commit()

    return response


async def delete_response(db: "AsyncSession", search_engine: SearchEngine, response: Response) -> Response:
    async with db.begin_nested():
        response = await response.delete(db, autocommit=False)
        await search_engine.delete_record_response(response)

    await db.commit()

    return response


def validate_response_values(dataset: Dataset, values: Dict[str, ResponseValue], status: ResponseStatus):
    if not values:
        if status != ResponseStatus.discarded:
            raise ValueError("Missing response values")
        return

    values_copy = copy.copy(values or {})
    for question in dataset.questions:
        if (
            question.required
            and status == ResponseStatus.submitted
            and not (question.name in values and values_copy.get(question.name))
        ):
            raise ValueError(f"Missing required question: {question.name!r}")

        question_response = values_copy.pop(question.name, None)
        if question_response:
            question.parsed_settings.check_response(question_response, status)

    if values_copy:
        raise ValueError(f"Error: found responses for non configured questions: {list(values_copy.keys())!r}")


def validate_record_fields(dataset: Dataset, fields: Dict[str, Any]):
    fields_copy = copy.copy(fields or {})
    for field in dataset.fields:
        if field.required and not (field.name in fields_copy and fields_copy.get(field.name) is not None):
            raise ValueError(f"Missing required value for field: {field.name!r}")

        value = fields_copy.pop(field.name, None)
        if not isinstance(value, str):
            raise ValueError(
                f"Wrong value found for field {field.name!r}. Expected {str.__name__!r}, found {type(value).__name__!r}"
            )

    if fields_copy:
        raise ValueError(f"Error: found fields values for non configured fields: {list(fields_copy.keys())!r}")


async def get_suggestion_by_record_id_and_question_id(
    db: "AsyncSession", record_id: UUID, question_id: UUID
) -> Union[Suggestion, None]:
    result = await db.execute(select(Suggestion).filter_by(record_id=record_id, question_id=question_id))
    return result.scalar_one_or_none()


async def upsert_suggestion(
    db: "AsyncSession", record: Record, question: Question, suggestion_create: "SuggestionCreate"
) -> Suggestion:
    question.parsed_settings.check_response(suggestion_create)
    return await Suggestion.upsert(
        db,
        schema=SuggestionCreateWithRecordId(record_id=record.id, **suggestion_create.dict()),
        constraints=[Suggestion.record_id, Suggestion.question_id],
    )


async def delete_suggestions(db: "AsyncSession", record: Record, suggestions_ids: List[UUID]) -> None:
    params = [Suggestion.id.in_(suggestions_ids), Suggestion.record_id == record.id]
    await Suggestion.delete_many(db=db, params=params)


async def get_suggestion_by_id(db: "AsyncSession", suggestion_id: "UUID") -> Union[Suggestion, None]:
    result = await db.execute(
        select(Suggestion)
        .filter_by(id=suggestion_id)
        .options(selectinload(Suggestion.record).selectinload(Record.dataset))
    )
    return result.scalar_one_or_none()


async def delete_suggestion(db: "AsyncSession", suggestion: Suggestion) -> Suggestion:
    return await suggestion.delete(db)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, List, Optional, Tuple

from fastapi import Depends

from argilla.server.errors.base_errors import EntityAlreadyExistsError, EntityNotFoundError, MissingDatasetRecordsError
from argilla.server.services.datasets import DatasetsService
from argilla.server.services.metrics import MetricsService
from argilla.server.services.search.model import ServiceSearchResults, ServiceSortableField, ServiceSortConfig
from argilla.server.services.search.service import SearchRecordsService
from argilla.server.services.storage.service import RecordsStorageService
from argilla.server.services.tasks.commons import BulkResponse
from argilla.server.services.tasks.text_classification.metrics import TextClassificationMetrics
from argilla.server.services.tasks.text_classification.model import (
    DatasetLabelingRulesMetricsSummary,
    DatasetLabelingRulesSummary,
    LabelingRuleMetricsSummary,
    LabelingRuleSummary,
    ServiceLabelingRule,
    ServiceTextClassificationDataset,
    ServiceTextClassificationQuery,
    ServiceTextClassificationRecord,
)


class TextClassificationService:
    """
    Text classification service

    """

    _INSTANCE = None

    @classmethod
    def get_instance(
        cls,
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        metrics: MetricsService = Depends(MetricsService.get_instance),
        storage: RecordsStorageService = Depends(RecordsStorageService.get_instance),
        search: SearchRecordsService = Depends(SearchRecordsService.get_instance),
    ) -> "TextClassificationService":
        if not cls._INSTANCE:
            cls._INSTANCE = cls(datasets=datasets, metrics=metrics, storage=storage, search=search)
        return cls._INSTANCE

    def __init__(
        self,
        datasets: DatasetsService,
        metrics: MetricsService,
        storage: RecordsStorageService,
        search: SearchRecordsService,
    ):
        self.__storage__ = storage
        self.__search__ = search
        self.__metrics__ = metrics
        self.__datasets__ = datasets

    async def add_records(
        self,
        dataset: ServiceTextClassificationDataset,
        records: List[ServiceTextClassificationRecord],
    ):
        if not records:
            return BulkResponse(dataset=dataset.name, processed=0)

        # TODO(@frascuchon): This will moved to dataset settings validation once DatasetSettings join the game!
        self._check_multi_label_integrity(dataset, records)

        failed = await self.__storage__.store_records(
            dataset=dataset,
            records=records,
            record_type=ServiceTextClassificationRecord,
        )
        return BulkResponse(dataset=dataset.name, processed=len(records), failed=failed)

    def search(
        self,
        dataset: ServiceTextClassificationDataset,
        query: ServiceTextClassificationQuery,
        sort_by: List[ServiceSortableField],
        record_from: int = 0,
        size: int = 100,
        exclude_metrics: bool = True,
    ) -> ServiceSearchResults:
        """
        Run a search in a dataset

        Parameters
        ----------
        dataset:
            The records dataset
        query:
            The search parameters
        sort_by:
            The sort by list
        record_from:
            The record from return results
        size:
            The max number of records to return
        Returns
        -------
            The matched records with aggregation info for specified task_meta.py

        """

        metrics = [
            TextClassificationMetrics.find_metric(id)
            for id in {
                "words_cloud",
                "predicted_by",
                "predicted_as",
                "annotated_by",
                "annotated_as",
                "error_distribution",
                "status_distribution",
                "metadata",
                "score",
            }
        ]

        results = self.__search__.search(
            dataset,
            query=query,
            record_type=ServiceTextClassificationRecord,
            record_from=record_from,
            size=size,
            exclude_metrics=exclude_metrics,
            metrics=metrics,
            sort_config=ServiceSortConfig(
                sort_by=sort_by,
            ),
        )

        if results.metrics:
            results.metrics["words"] = results.metrics.get("words_cloud", {})
            results.metrics["status"] = results.metrics["status_distribution"]
            results.metrics["predicted"] = results.metrics["error_distribution"]
            results.metrics["predicted"].pop("unknown", None)

        return results

    def read_dataset(
        self,
        dataset: ServiceTextClassificationDataset,
        query: Optional[ServiceTextClassificationQuery] = None,
        id_from: Optional[str] = None,
        limit: int = 1000,
    ) -> Iterable[ServiceTextClassificationRecord]:
        """
        Scan a dataset records

        Parameters
        ----------
        dataset:
            The dataset name
        query:
            If provided, scan will retrieve only records matching
            the provided query filters. Optional
        id_from:
            If provided, read the samples after this record ID
        limit:
            Batch size to scan, only used if `id_from` is specified

        """
        yield from self.__search__.scan_records(
            dataset,
            query=query,
            record_type=ServiceTextClassificationRecord,
            id_from=id_from,
            limit=limit,
        )

    def _check_multi_label_integrity(
        self,
        dataset: ServiceTextClassificationDataset,
        records: List[ServiceTextClassificationRecord],
    ):
        is_multi_label_dataset = self._is_dataset_multi_label(dataset)
        if is_multi_label_dataset is not None:
            is_multi_label = records[0].multi_label
            assert (
                is_multi_label == is_multi_label_dataset
            ), "You cannot pass {labels_type} records for this dataset. " "Stored records are {labels_type}".format(
                labels_type="multi-label" if is_multi_label else "single-label"
            )

    def _is_dataset_multi_label(self, dataset: ServiceTextClassificationDataset) -> Optional[bool]:
        try:
            results = self.__search__.search(
                dataset,
                record_type=ServiceTextClassificationRecord,
                size=1,
            )
        except MissingDatasetRecordsError:  # No records index yet
            return None
        if results.records:
            return results.records[0].multi_label

    def find_labeling_rule(
        self, dataset: ServiceTextClassificationDataset, rule_query: str, error_on_missing: bool = True
    ) -> Optional[ServiceLabelingRule]:
        rule_query = rule_query.strip()
        for rule in dataset.rules:
            if rule.query == rule_query:
                return rule
        if error_on_missing:
            raise EntityNotFoundError(rule_query, type=ServiceLabelingRule)

    def list_labeling_rules(self, dataset: ServiceTextClassificationDataset) -> Iterable[ServiceLabelingRule]:
        return dataset.rules

    def add_labeling_rule(self, dataset: ServiceTextClassificationDataset, rule: ServiceLabelingRule) -> None:
        """
        Adds a labeling rule

        Parameters
        ----------
        dataset:
            The dataset

        rule:
            The rule
        """

        self.__normalized_rule__(rule)
        if self.find_labeling_rule(dataset, rule_query=rule.query, error_on_missing=False):
            raise EntityAlreadyExistsError(rule.query, type=ServiceLabelingRule)

        dataset.rules.append(rule)
        self.__datasets__.raw_dataset_update(dataset)

    def update_labeling_rule(
        self,
        dataset: ServiceTextClassificationDataset,
        rule_query: str,
        labels: List[str],
        description: Optional[str] = None,
    ) -> ServiceLabelingRule:
        found_rule = self.find_labeling_rule(dataset, rule_query)

        found_rule.labels = labels
        found_rule.label = labels[0] if len(labels) == 1 else None
        if description is not None:
            found_rule.description = description

        self.__normalized_rule__(found_rule)
        for idx, r in enumerate(dataset.rules):
            if r.query == found_rule.query:
                dataset.rules[idx] = found_rule
                break
        self.__datasets__.raw_dataset_update(dataset)

        return found_rule

    def delete_labeling_rule(self, dataset: ServiceTextClassificationDataset, rule_query: str):
        """Delete a rule from a dataset by its defined query string"""
        new_rules_set = [r for r in dataset.rules if r.query != rule_query]
        if len(dataset.rules) != new_rules_set:
            dataset.rules = new_rules_set
            self.__datasets__.raw_dataset_update(dataset)

    def compute_labeling_rule(
        self,
        dataset: ServiceTextClassificationDataset,
        rule_query: str,
        labels: Optional[List[str]] = None,
    ) -> LabelingRuleMetricsSummary:
        """
        Compute metrics for a given rule. It's not necessary that query rule
        is created in dataset. Basic computed rules are:

        coverage, correct[*], incorrect[*], precision[*]

        [*]: computed metrics only if label is provided or rule already created in dataset

        Parameters
        ----------
        dataset:
            The dataset
        rule_query:
            The provided rule query. If already created in dataset, the ``label``
            param will be omitted
        labels:
            Label used for the rule metrics. If not provided and no rule was stored with the
            provided query, no precision will be computed.
            Otherwise, the labels from the stored rule will be used to compute the metrics.

        Returns
        -------

            Metrics summary for rule and labels

        """

        rule_query = rule_query.strip()

        if labels is None:
            rule = self.find_labeling_rule(dataset, rule_query=rule_query, error_on_missing=False)
            if rule:
                labels = rule.labels

        metric_data = self.__metrics__.summarize_metric(
            dataset=dataset,
            metric=TextClassificationMetrics.find_metric("labeling_rule"),
            rule_query=rule_query,
            labels=labels,
        )
        annotated = self.__metrics__.annotated_records(dataset)
        total = self.__metrics__.total_records(dataset)

        metrics = LabelingRuleSummary.parse_obj(metric_data)

        coverage = metrics.covered_records / total if total > 0 else None
        coverage_annotated = metrics.annotated_covered_records / annotated if annotated > 0 else None

        return LabelingRuleMetricsSummary(
            total_records=total,
            annotated_records=annotated,
            coverage=coverage,
            coverage_annotated=coverage_annotated,
            correct=metrics.correct_records if annotated > 0 else None,
            incorrect=metrics.incorrect_records if annotated > 0 else None,
            precision=metrics.precision if annotated > 0 else None,
        )

    def compute_all_labeling_rules(self, dataset: ServiceTextClassificationDataset):
        total, annotated, metrics = self._compute_all_lb_rules_metrics(dataset)
        coverage = metrics.covered_records / total if total else None
        coverage_annotated = metrics.annotated_covered_records / annotated if annotated else None
        return DatasetLabelingRulesMetricsSummary(
            coverage=coverage,
            coverage_annotated=coverage_annotated,
            total_records=total,
            annotated_records=annotated,
        )

    def _compute_all_lb_rules_metrics(
        self, dataset: ServiceTextClassificationDataset
    ) -> Tuple[int, int, DatasetLabelingRulesSummary]:
        annotated_records = self.__metrics__.annotated_records(dataset)
        dataset_records = self.__metrics__.total_records(dataset)
        metric_data = self.__metrics__.summarize_metric(
            dataset=dataset,
            metric=TextClassificationMetrics.find_metric(id="dataset_labeling_rules"),
            queries=[r.query for r in dataset.rules],
        )

        return (
            dataset_records,
            annotated_records,
            DatasetLabelingRulesSummary.parse_obj(metric_data),
        )

    @staticmethod
    def __normalized_rule__(rule: ServiceLabelingRule) -> ServiceLabelingRule:
        if rule.labels and len(rule.labels) == 1:
            rule.label = rule.labels[0]
        elif rule.label and not rule.labels:
            rule.labels = [rule.label]

        return rule



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import textwrap
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Union

from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.training.schemas import TrainingTaskForTextClassification, TrainingTaskTypes
from argilla.client.models import Framework, TextClassificationRecord
from argilla.training import ArgillaTrainer as ArgillaTrainerV1

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

if TYPE_CHECKING:
    import spacy

    from argilla.client.feedback.dataset import FeedbackDataset


class ArgillaTrainer(ArgillaTrainerV1):
    def __init__(
        self,
        dataset: "FeedbackDataset",
        task: TrainingTaskTypes,
        framework: Framework,
        lang: Optional["spacy.Language"] = None,
        model: Optional[str] = None,
        train_size: Optional[float] = None,
        seed: Optional[int] = None,
        gpu_id: Optional[int] = -1,
        framework_kwargs: Optional[dict] = {},
    ) -> None:
        """
        Initialize an Argilla Trainer.

        Args:
            dataset: the dataset to be used for training.
            task: the training data to be used for training.
            framework: the framework to use for training. Currently, "transformers", "setfit", "spacy", "peft",
                "openai", "span_marker" and "trl" are supported.
            lang: the spaCy language model to use for training, just required when `framework="spacy"`.
                Defaults to None, but it will be set to `spacy.blank("en")` if not specified.
            model: name or path to the baseline model to be used. If not specified will set to a good default
                per framework, if applicable. Defaults to None.
            train_size: the size of the training set. If not specified, the entire dataset will be used for training,
                which may be an issue if `framework="spacy"` as it requires a validation set. Defaults to None.
            seed: the random seed to ensure reproducibility. Defaults to None.
            gpu_id: the GPU ID to use when training a SpaCy model. Defaults to -1, which means that the CPU
                will be used by default. GPU IDs start in 0, which stands for the default GPU in the system,
                if available.
            framework_kwargs: arguments for the framework's trainer.
            **load_kwargs: arguments for the rg.load() function.
        """
        self._dataset = dataset
        self._train_size = train_size
        self._task = task
        self._seed = seed  # split is used for train-test-split and should therefore be fixed
        self._model = model

        self._prepared_data = self._dataset.prepare_for_training(
            framework=framework,
            task=task,
            train_size=train_size,
            seed=seed,
            lang=lang,
        )

        if isinstance(framework, str):
            framework = Framework(framework)

        if framework is Framework.SETFIT:
            if not isinstance(task, TrainingTaskForTextClassification):
                raise NotImplementedError(f"{Framework.SETFIT} only supports `TextClassification` tasks.")
            from argilla.client.feedback.training.frameworks.setfit import ArgillaSetFitTrainer

            self._trainer = ArgillaSetFitTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.transformers import ArgillaTransformersTrainer

            self._trainer = ArgillaTransformersTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.PEFT:
            from argilla.client.feedback.training.frameworks.peft import ArgillaPeftTrainer

            self._trainer = ArgillaPeftTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.SPACY:
            from argilla.client.feedback.training.frameworks.spacy import ArgillaSpaCyTrainer

            self._trainer = ArgillaSpaCyTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                gpu_id=gpu_id,
                framework_kwargs=framework_kwargs,  # freeze_tok2vec
            )
        elif framework is Framework.SPACY_TRANSFORMERS:
            from argilla.client.feedback.training.frameworks.spacy import ArgillaSpaCyTransformersTrainer

            self._trainer = ArgillaSpaCyTransformersTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
                gpu_id=gpu_id,
                framework_kwargs=framework_kwargs,  # update_transformer
            )
        elif framework is Framework.OPENAI:
            from argilla.client.feedback.training.frameworks.openai import ArgillaOpenAITrainer

            self._trainer = ArgillaOpenAITrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.SPAN_MARKER:
            from argilla.client.feedback.training.frameworks.span_marker import ArgillaSpanMarkerTrainer

            self._trainer = ArgillaSpanMarkerTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        elif framework is Framework.TRL:
            from argilla.client.feedback.training.frameworks.trl import ArgillaTRLTrainer

            self._trainer = ArgillaTRLTrainer(
                dataset=self._dataset,
                task=self._task,
                prepared_data=self._prepared_data,
                seed=self._seed,
                model=self._model,
            )
        else:
            raise NotImplementedError(f"{framework} is not a valid framework.")

        self._logger.info(self)
        self._track_trainer_usage(framework=framework, task=self._task.__class__.__name__)

    def __repr__(self) -> str:
        """
        `trainer.__repr__()` prints out the trainer's parameters and a summary of how to use the trainer

        Returns:
          The trainer object.
        """
        return textwrap.dedent(
            f"""\
            ArgillaBaseTrainer info:
            _________________________________________________________________
            These baseline params are fixed:
                dataset: {self._dataset}
                task: {self._task}
                train_size: {self._train_size}
                seed: {self._seed}


            {self._trainer.__class__} info:
            _________________________________________________________________
            The parameters are configurable via `trainer.update_config()`:
                {self._trainer}

            Using the trainer:
            _________________________________________________________________
            `trainer.train(output_dir)` to train to start training. `output_dir` is the directory to save the model automatically.
            `trainer.predict(text, as_argilla_records=True)` to make predictions.
            `trainer.save(output_dir)` to save the model manually.
            """
        )

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        `predict` takes a string or list of strings and returns a list of dictionaries, each dictionary
        containing the text, the predicted label, and the confidence score.

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the output will be a list of Argilla records instead of dictionaries. Defaults to True.

        Returns:
          A list of predictions or Argilla records.
        """
        return self._trainer.predict(text=text, as_argilla_records=False, **kwargs)


class ArgillaTrainerSkeleton(ABC):
    def __init__(
        self,
        dataset: "FeedbackDataset",
        task: TrainingTaskTypes,
        prepared_data=None,
        model: str = None,
        seed: int = None,
        *arg,
        **kwargs,
    ):
        self._dataset = dataset
        self._task = task
        self._dataset = prepared_data
        self._model = model
        self._seed = seed
        if isinstance(self._task, TrainingTaskForTextClassification):
            self._multi_label = self._task.__multi_label__ or False
            self._label_list = self._task.__all_labels__ or None
            self._label2id = self._task.__label2id__
            self._id2label = self._task.__id2label__
            self._record_class = TextClassificationRecord  # TODO: dirty hack to inherit from original trainers
        else:
            self._record_class = FeedbackRecord

    @abstractmethod
    def init_training_args(self) -> None:
        """
        Initializes the training arguments.
        """

    @abstractmethod
    def init_model(self) -> None:
        """
        Initializes a model.
        """

    @abstractmethod
    def update_config(self, *args, **kwargs) -> None:
        """
        Updates the configuration of the trainer, but the parameters depend on the trainer.subclass.
        """

    @abstractmethod
    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        Predicts the label of the text.
        """

    @abstractmethod
    def train(self, output_dir: Optional[str] = None) -> None:
        """
        Trains the model.
        """

    @abstractmethod
    def save(self, output_dir: str) -> None:
        """
        Saves the model to the specified path.
        """



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Union
from uuid import UUID

import httpx

from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.v1.workspaces.models import WorkspaceModel


def get_workspace(client: httpx.Client, id: UUID) -> Response[Union[WorkspaceModel, ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/workspaces/{id}` endpoint to retrieve a workspace.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the workspace to be retrieved.

    Returns:
        A `Response` object with the parsed response, containing a `parsed` attribute with the
        parsed response if the request was successful, which is an instance of `WorkspaceModel`.
    """
    url = f"/api/v1/workspaces/{id}"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = WorkspaceModel(**response.json())
        return response_obj
    return handle_response_error(response)


def delete_workspace(
    client: httpx.Client, id: UUID
) -> Response[Union[WorkspaceModel, ErrorMessage, HTTPValidationError]]:
    url = f"/api/v1/workspaces/{id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = WorkspaceModel(**response.json())
        return response_obj
    return handle_response_error(response)


def list_workspaces_me(
    client: httpx.Client,
) -> Response[Union[List[WorkspaceModel], ErrorMessage, HTTPValidationError]]:
    """Sends a GET request to `/api/v1/me/workspaces` endpoint to get the list of
    workspaces the current user has access to.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a list of `WorkspaceModel`.
    """
    url = "/api/v1/me/workspaces"

    response = client.get(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = [WorkspaceModel(**workspace) for workspace in response.json()["items"]]
        return response_obj
    return handle_response_error(response)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List, Optional, Tuple


class SpanUtils:
    """Holds utility methods to work with a tokenized text and entity spans.

    Spans must be tuples containing the label (str), start char idx (int), and end char idx (int).

    Args:
        text: The text the spans refer to.
        tokens: The tokens of the text.
    """

    def __init__(self, text: str, tokens: List[str]):
        self._text, self._tokens = text, tokens

        self._token_to_char_idx: Dict[int, Tuple[int, int]] = {}
        self._start_to_token_idx: Dict[int, int] = {}
        self._end_to_token_idx: Dict[int, int] = {}
        self._char_to_token_idx: Dict[int, int] = {}

        end_idx = 0
        for idx, token in enumerate(tokens):
            start_idx = text.find(token, end_idx)
            if start_idx == -1:
                raise ValueError(f"Token '{token}' not found in text: {text}")
            end_idx = start_idx + len(token)

            self._token_to_char_idx[idx] = (start_idx, end_idx)
            self._start_to_token_idx[start_idx] = idx
            self._end_to_token_idx[end_idx] = idx
            for i in range(start_idx, end_idx):
                self._char_to_token_idx[i] = idx

            # convention: skip first white space after a token
            try:
                if text[end_idx] == " ":
                    end_idx += 1
            # reached end of text
            except IndexError:
                pass

    @property
    def text(self) -> str:
        """The text the spans refer to."""
        return self._text

    @property
    def tokens(self) -> List[str]:
        """The tokens of the text."""
        return self._tokens

    @property
    def token_to_char_idx(self) -> Dict[int, Tuple[int, int]]:
        """The token index to start/end char index mapping."""
        return self._token_to_char_idx

    @property
    def char_to_token_idx(self) -> Dict[int, int]:
        """The char index to token index mapping."""
        return self._char_to_token_idx

    def validate(self, spans: List[Tuple[str, int, int]]):
        """Validates the alignment of span boundaries and tokens.

        Args:
            spans: A list of spans.

        Raises:
            ValueError: If a span is invalid, or if a span is not aligned with the tokens.
        """
        not_valid_spans_errors, misaligned_spans_errors = [], []

        for span in spans:
            char_start, char_end = span[1], span[2]
            if char_end - char_start < 1:
                not_valid_spans_errors.append(span)
            elif None in (
                self._start_to_token_idx.get(char_start),
                self._end_to_token_idx.get(char_end),
            ):
                span_str = self.text[char_start:char_end]
                message = f"{span} - {repr(span_str)}"
                misaligned_spans_errors.append(message)

        if not_valid_spans_errors or misaligned_spans_errors:
            message = ""
            if not_valid_spans_errors:
                message += f"Following entity spans are not valid: {not_valid_spans_errors}\n"

            if misaligned_spans_errors:
                spans = "\n".join(misaligned_spans_errors)
                message += "Following entity spans are not aligned with provided tokenization\n"
                message += f"Spans:\n{spans}\n"
                message += f"Tokens:\n{self.tokens}"

            raise ValueError(message)

    def correct(self, spans: List[Tuple[str, int, int]]) -> List[Tuple[str, int, int]]:
        """Correct span boundaries for leading/trailing white spaces, new lines and tabs.

        Args:
            spans: Spans to be corrected.

        Returns:
            The corrected spans.
        """
        corrected_spans = []
        for span in spans:
            start, end = span[1], span[2]

            if start < 0:
                start = 0
            if end > len(self.text):
                end = len(self.text)

            while start <= len(self.text) and not self.text[start].strip():
                start += 1
            while not self.text[end - 1].strip():
                end -= 1

            corrected_spans.append((span[0], start, end))

        return corrected_spans

    def to_tags(self, spans: List[Tuple[str, int, int]]) -> List[str]:
        """Convert spans to IOB tags.

        Args:
            spans: Spans to transform into IOB tags.

        Returns:
            The IOB tags.

        Raises:
            ValueError: If spans overlap, the IOB format does not support overlapping spans.
        """
        # check for overlapping spans
        sorted_spans = sorted(spans, key=lambda x: x[1])
        for i in range(1, len(spans)):
            if sorted_spans[i - 1][2] > sorted_spans[i][1]:
                raise ValueError("IOB tags cannot handle overlapping spans!")

        tags = ["O"] * len(self.tokens)
        for span in spans:
            start_token_idx = self._start_to_token_idx[span[1]]
            end_token_idx = self._end_to_token_idx[span[2]]

            tags[start_token_idx] = f"B-{span[0]}"
            for token_idx in range(start_token_idx + 1, end_token_idx + 1):
                tags[token_idx] = f"I-{span[0]}"

        return tags

    def from_tags(self, tags: List[str]) -> List[Tuple[str, int, int]]:
        """Convert IOB or BILOU tags to spans.

        Overlapping spans are NOT supported!

        Args:
            tags: The IOB or BILOU tags.

        Returns:
            A list of spans.

        Raises:
            ValueError: If the list of tags has not the same length as the list of tokens,
                or tags are not in the IOB or BILOU format.
        """

        def get_prefix_and_entity(tag_str: str) -> Tuple[str, Optional[str]]:
            if tag_str == "O":
                return tag_str, None
            splits = tag_str.split("-")
            return splits[0], "-".join(splits[1:])

        if len(tags) != len(self.tokens):
            raise ValueError("The list of tags must have the same length as the list of tokens!")

        spans, start_idx = [], None
        for idx, tag in enumerate(tags):
            prefix, entity = get_prefix_and_entity(tag)

            if prefix == "O":
                continue

            if prefix == "U":
                start_idx, end_idx = self._token_to_char_idx[idx]
                spans.append((entity, start_idx, end_idx))
                start_idx = None
                continue

            if prefix == "L":
                # If no start prefix, we just assume "L" == "U":
                if start_idx is None:
                    start_idx, end_idx = self._token_to_char_idx[idx]
                else:
                    _, end_idx = self._token_to_char_idx[idx]
                spans.append((entity, start_idx, end_idx))
                start_idx = None
                continue

            if prefix == "B":
                start_idx, end_idx = self._token_to_char_idx[idx]
            elif prefix == "I":
                # If "B" is missing, we just assume "I" starts the span
                if start_idx is None:
                    start_idx = self._token_to_char_idx[idx][0]
                end_idx = self._token_to_char_idx[idx][1]
            else:
                raise ValueError("Tags are not in the IOB or BILOU format!")

            try:
                next_tag = tags[idx + 1]
            # Reached last tag, add span
            except IndexError:
                spans.append((entity, start_idx, end_idx))
                break

            next_prefix, next_entity = get_prefix_and_entity(next_tag)
            # span continues
            if next_prefix in ["I", "L"] and next_entity == entity:
                continue
            # span ends
            spans.append((entity, start_idx, end_idx))
            start_idx = None

        return spans



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TYPE_CHECKING, Dict, List, Union

from argilla.client.feedback.training.base import ArgillaTrainerSkeleton
from argilla.client.feedback.training.schemas import (
    TrainingTaskForDPO,
    TrainingTaskForPPO,
    TrainingTaskForRM,
    TrainingTaskForSFT,
)
from argilla.training.utils import filter_allowed_args
from argilla.utils.dependency import require_version

if TYPE_CHECKING:
    import transformers
    from trl import PPOConfig

    from argilla.client.feedback.dataset import FeedbackDataset


class PPOArgs:
    def __init__(
        self,
        config: "PPOConfig",
        reward_model: Union[str, "transformers.pipeline"],
        length_sampler_kwargs: dict,
        generation_kwargs: dict,
    ) -> None:
        """
        Additional arguments for PPO training process.

        Args:
            reward_model (Union[str, "transformers.pipeline"]): Reward model to use for creating PPO rewards for training.
            length_sampler_kwargs (dict): Arguments for the length sampler.
                min_value: Minimum length for the generated samples.
                max_value: Maximum length for the generated samples.
            generation_kwargs (dict): Arguments for the generation process.
                min_length: Minimum length for the generated samples.
                max_length: Maximum length for the generated samples.
                num_beams: Number of beams to use for the generation process.
                num_return_sequences: Number of sequences to generate.
                temperature: Temperature for the generation process.
                top_k: Top k for the generation process.
                top_p: Top p for the generation process.
        """
        if isinstance(reward_model, str):
            from transformers import pipeline

            reward_model = pipeline("text-classifcation", model=reward_model)
        self.config = config
        self.reward_model = reward_model
        self.length_sampler_kwargs = length_sampler_kwargs
        self.generation_kwargs = generation_kwargs


class ArgillaTRLTrainer(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaTRLTrainer")
    _logger.setLevel(logging.INFO)

    require_version("transformers")
    require_version("torch")
    require_version("trl>=0.5.0")

    def __init__(
        self,
        dataset: "FeedbackDataset",
        task: Union[
            TrainingTaskForSFT,
            TrainingTaskForRM,
            TrainingTaskForPPO,
            TrainingTaskForDPO,
        ],
        prepared_data=None,
        model: str = None,
        seed: int = None,
    ) -> None:
        super().__init__(dataset=dataset, task=task, prepared_data=prepared_data, model=model, seed=seed)
        import torch
        from datasets import DatasetDict
        from transformers import set_seed

        self._transformers_model = None
        self._transformers_tokenizer = None
        self.device = "cpu"
        if torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda"

        if self._seed is None:
            self._seed = 42
        set_seed(self._seed)

        if self._model is None:
            self._model = "gpt2-medium"

        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None

        if not isinstance(
            self._task,
            (
                TrainingTaskForSFT,
                TrainingTaskForRM,
                TrainingTaskForPPO,
                TrainingTaskForDPO,
            ),
        ):
            raise NotImplementedError(f"Task {self._task} not supported in TRL.")

        from trl import DPOTrainer, PPOTrainer, RewardTrainer, SFTTrainer

        self.trainer_mapping = {
            TrainingTaskForSFT: SFTTrainer,
            TrainingTaskForRM: RewardTrainer,
            TrainingTaskForPPO: PPOTrainer,
            TrainingTaskForDPO: DPOTrainer,
        }
        self.trainer_cls = self.trainer_mapping[type(self._task)]

        self.init_training_args()

    def init_training_args(self) -> None:
        """
        Initializes the training arguments.
        """
        self.training_args_kwargs = {}
        self.trainer_kwargs = {}

        if isinstance(self._task, TrainingTaskForPPO):
            from trl import PPOConfig

            self._logger.warning(
                "The PPOTrainer must be initialized by passing `reward_model`, `length_sampler_kwargs`, `generation_kwargs` as kwargs to the `update_config()`-method."
            )
            self.trainer_kwargs["config"] = PPOConfig()
            self.training_args_kwargs["reward_model"] = None
            self.training_args_kwargs["length_sampler_kwargs"] = {"min_value": 1, "max_value": 10}
            self.training_args_kwargs["generation_kwargs"] = {
                "min_length": -1,
                "top_k": 0.0,
                "top_p": 1.0,
                "do_sample": True,
            }
        else:
            self.training_args_kwargs["evaluation_strategy"] = "no" if self._eval_dataset is None else "epoch"
            self.training_args_kwargs["logging_steps"] = 30
            self.training_args_kwargs["logging_steps"] = 1
            self.training_args_kwargs["num_train_epochs"] = 1

    def init_model(self, new: bool = False) -> None:
        """
        Initializes a model.
        """
        from transformers import (
            AutoModelForCausalLM,
            AutoModelForSequenceClassification,
            AutoTokenizer,
            PreTrainedModel,
            PreTrainedTokenizer,
        )
        from trl import AutoModelForCausalLMWithValueHead, create_reference_model

        if isinstance(self._task, (TrainingTaskForSFT, TrainingTaskForDPO)):
            auto_model_class = AutoModelForCausalLM
        elif isinstance(self._task, TrainingTaskForPPO):
            auto_model_class = AutoModelForCausalLMWithValueHead
        elif isinstance(self._task, TrainingTaskForRM):
            auto_model_class = AutoModelForSequenceClassification

        self._transformers_model: PreTrainedModel = auto_model_class.from_pretrained(self._model)
        self._transformers_tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(self._model)
        self._transformers_tokenizer.pad_token = self._transformers_tokenizer.eos_token
        self._transformers_model.config.pad_token_id = self._transformers_tokenizer.pad_token_id

        if isinstance(self._task, (TrainingTaskForPPO, TrainingTaskForDPO)):
            self._transformers_ref_model: PreTrainedModel = create_reference_model(self._transformers_model)
        if new:
            self._transformers_model.to(self.device)

    def update_config(self, **kwargs) -> None:
        """
        Updates the configuration of the trainer, but the parameters depend on the trainer.subclass.
        """
        from transformers import TrainingArguments

        self.training_args_kwargs.update(filter_allowed_args(TrainingArguments.__init__, **kwargs))
        self.training_args_kwargs.update(filter_allowed_args(PPOArgs.__init__, **kwargs))
        self.trainer_kwargs.update(filter_allowed_args(self.trainer_cls.__init__, **kwargs))

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs) -> None:
        """
        Predicts the label of the text.
        """
        raise NotImplementedError("Models trained with TRL cannot be used for label predictions.")

    def train(self, output_dir: str) -> None:
        """
        Trains the model.
        """
        if isinstance(self._task, TrainingTaskForPPO):
            if not all(
                x in self.training_args_kwargs for x in ["length_sampler_kwargs", "generation_kwargs", "reward_model"]
            ):
                raise ValueError(
                    "To train a PPO model, you need to specify the following arguments via `trainer.update_config`: length_sampler_kwargs, generation_kwargs, reward_model."
                )

        from transformers import TrainingArguments

        # check required path argument
        self.training_args_kwargs["output_dir"] = output_dir

        self.init_model(new=True)

        if isinstance(self._task, TrainingTaskForSFT):
            self._training_args = TrainingArguments(**self.training_args_kwargs)
            self._trainer = self.trainer_cls(
                self._transformers_model,
                args=self._training_args,
                train_dataset=self._train_dataset,
                eval_dataset=self._eval_dataset,
                dataset_text_field="text",
                tokenizer=self._transformers_tokenizer,
                **self.trainer_kwargs,
            )

        elif isinstance(self._task, TrainingTaskForRM):

            def preprocess_function(examples) -> Dict[str, List]:
                new_examples = {
                    "input_ids_chosen": [],
                    "attention_mask_chosen": [],
                    "input_ids_rejected": [],
                    "attention_mask_rejected": [],
                }
                for chosen, rejected in zip(examples["chosen"], examples["rejected"]):
                    tokenized_j = self._transformers_tokenizer(chosen, truncation=True)
                    tokenized_k = self._transformers_tokenizer(rejected, truncation=True)

                    new_examples["input_ids_chosen"].append(tokenized_j["input_ids"])
                    new_examples["attention_mask_chosen"].append(tokenized_j["attention_mask"])
                    new_examples["input_ids_rejected"].append(tokenized_k["input_ids"])
                    new_examples["attention_mask_rejected"].append(tokenized_k["attention_mask"])

                return new_examples

            self._training_args = TrainingArguments(**self.training_args_kwargs)
            train_dataset = self._train_dataset.map(preprocess_function, batched=True)
            eval_dataset = None
            if self._eval_dataset:
                eval_dataset = self._eval_dataset.map(preprocess_function, batched=True)

            self._trainer = self.trainer_cls(
                self._transformers_model,
                args=self._training_args,
                train_dataset=train_dataset,
                eval_dataset=eval_dataset,
                tokenizer=self._transformers_tokenizer,
                **self.trainer_kwargs,
            )
        elif isinstance(self._task, TrainingTaskForPPO):
            from datasets import concatenate_datasets

            dataset = concatenate_datasets([x for x in [self._train_dataset, self._eval_dataset] if x is not None])

            def tokenize(sample):
                sample["input_ids"] = self._transformers_tokenizer.encode(sample["query"], truncation=True)
                return sample

            def data_collator(data):
                return dict((key, [d[key] for d in data]) for key in data[0])

            def remove_truncated(sample):
                return len(sample) < self._transformers_tokenizer.model_max_length

            dataset = dataset.map(tokenize, batched=False)
            size_before = len(dataset)
            dataset = dataset.filter(remove_truncated, batched=False)
            size_after = len(dataset)
            if size_after != size_before:
                self._logger.info(
                    f"Removed {size_before - size_after} samples ({1 - (size_after / size_before):%}), "
                    "as these samples were longer than the maximum model length even before the generation."
                )
            dataset.set_format(type="torch")
            self._trainer = self.trainer_cls(
                model=self._transformers_model,
                ref_model=self._transformers_ref_model,
                tokenizer=self._transformers_tokenizer,
                dataset=dataset,
                data_collator=data_collator,
                **self.trainer_kwargs,
            )
        elif isinstance(self._task, TrainingTaskForDPO):
            self._training_args = TrainingArguments(**self.training_args_kwargs)
            self._trainer = self.trainer_cls(
                model=self._transformers_model,
                ref_model=self._transformers_ref_model,
                args=self._training_args,
                train_dataset=self._train_dataset,
                eval_dataset=self._eval_dataset,
                tokenizer=self._transformers_tokenizer,
                **self.trainer_kwargs,
            )

        #  train
        if isinstance(self._task, TrainingTaskForPPO):
            import torch
            from tqdm import tqdm
            from trl.core import LengthSampler

            output_length_sampler = LengthSampler(**self.training_args_kwargs["length_sampler_kwargs"])
            generation_kwargs = self.training_args_kwargs["generation_kwargs"]
            generation_kwargs["pad_token_id"] = self._transformers_tokenizer.eos_token_id
            reward_model = self.training_args_kwargs["reward_model"]

            for batch in tqdm(self._trainer.dataloader):
                query_tensors = batch["input_ids"]

                #### Get response from SFT
                response_tensors = self._trainer.generate(
                    query_tensors,
                    return_prompt=False,
                    length_sampler=output_length_sampler,
                    **generation_kwargs,
                )
                batch["response"] = self._transformers_tokenizer.batch_decode(
                    response_tensors, skip_special_tokens=True
                )

                #### Compute rewards scores
                texts = [q + r for q, r in zip(batch["query"], batch["response"])]
                pipe_outputs = reward_model(texts, top_k=None, truncation=True)
                rewards = [torch.tensor(output[-1]["score"] * len(output)) for output in pipe_outputs]

                #### Run PPO step
                stats = self._trainer.step(query_tensors, response_tensors, rewards)
                self._trainer.log_stats(stats, batch, rewards)
        else:
            self._trainer.train()
            if self._trainer.eval_dataset:
                self._metrics = self._trainer.evaluate()
                self._logger.info(self._metrics)
            else:
                self._metrics = None

        self.save(output_dir)

    def save(self, output_dir: str) -> None:
        """
        Saves the model to the specified path.
        """
        self._transformers_model.save_pretrained(output_dir)
        self._transformers_tokenizer.save_pretrained(output_dir)

    def __repr__(self) -> str:
        formatted_string = []
        arg_dict = {
            repr(self.trainer_cls.__name__): self.trainer_kwargs,
            "'TrainingArguments'": self.training_args_kwargs,
        }
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\n".join(formatted_string)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import re
import uuid
from datetime import datetime
from typing import Any, List, Optional
from uuid import UUID

from pydantic import BaseModel, Field, constr, root_validator, validator
from pydantic.utils import GetterDict

from argilla._constants import ES_INDEX_REGEX_PATTERN
from argilla.server.errors import BadRequestError, EntityNotFoundError
from argilla.server.models import UserRole

WORKSPACE_NAME_REGEX = ES_INDEX_REGEX_PATTERN

USER_USERNAME_REGEX = ES_INDEX_REGEX_PATTERN
USER_PASSWORD_MIN_LENGTH = 8
USER_PASSWORD_MAX_LENGTH = 100


class WorkspaceUserCreate(BaseModel):
    user_id: UUID
    workspace_id: UUID


class Workspace(BaseModel):
    id: UUID
    name: str
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class WorkspaceCreate(BaseModel):
    name: constr(regex=WORKSPACE_NAME_REGEX, min_length=1)


class UserCreate(BaseModel):
    first_name: constr(min_length=1, strip_whitespace=True)
    last_name: Optional[constr(min_length=1, strip_whitespace=True)]
    username: constr(regex=USER_USERNAME_REGEX, min_length=1)
    role: Optional[UserRole]
    password: constr(min_length=USER_PASSWORD_MIN_LENGTH, max_length=USER_PASSWORD_MAX_LENGTH)
    workspaces: Optional[List[str]]


class UserGetter(GetterDict):
    def get(self, key: str, default: Any = None) -> Any:
        if key == "full_name":
            return f"{self._obj.first_name} {self._obj.last_name}" if self._obj.last_name else self._obj.first_name
        elif key == "workspaces":
            return [workspace.name for workspace in self._obj.workspaces]
        else:
            return super().get(key, default)


class User(BaseModel):
    """Base user model"""

    id: UUID
    first_name: str
    last_name: Optional[str]
    full_name: Optional[str] = Field(description="Deprecated. Use `first_name` and `last_name` instead")
    username: str = Field()
    role: UserRole
    workspaces: Optional[List[str]]
    api_key: str
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True
        getter_dict = UserGetter


class Token(BaseModel):
    """Token response model"""

    access_token: str
    token_type: str = "bearer"



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from tqdm import trange

from argilla.client.feedback.constants import DELETE_DATASET_RECORDS_MAX_NUMBER, PUSHING_BATCH_SIZE
from argilla.client.feedback.dataset.remote.base import RemoteFeedbackDatasetBase, RemoteFeedbackRecordsBase
from argilla.client.feedback.dataset.remote.filtered import FilteredRemoteFeedbackDataset
from argilla.client.feedback.schemas.records import FeedbackRecord, RemoteFeedbackRecord
from argilla.client.sdk.users.models import UserRole
from argilla.client.sdk.v1.datasets import api as datasets_api_v1
from argilla.client.sdk.v1.datasets.models import FeedbackResponseStatusFilter
from argilla.client.utils import allowed_for_roles

if TYPE_CHECKING:
    from uuid import UUID

    import httpx

    from argilla.client.feedback.schemas.types import AllowedFieldTypes, AllowedQuestionTypes
    from argilla.client.sdk.v1.datasets.models import FeedbackRecordsModel
    from argilla.client.workspaces import Workspace


class RemoteFeedbackRecords(RemoteFeedbackRecordsBase):
    def __init__(self, dataset: "RemoteFeedbackDataset") -> None:
        super().__init__(dataset=dataset)

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def __len__(self) -> int:
        """Returns the number of records in the current `FeedbackDataset` in Argilla."""
        try:
            response = datasets_api_v1.get_metrics(client=self._client, id=self._dataset.id)
        except Exception as e:
            raise Exception(
                f"Failed while getting the metrics from the current `FeedbackDataset` in Argilla with exception: {e}"
            ) from e
        return response.parsed.records.count

    def _fetch_records(self, offset: int, limit: int) -> "FeedbackRecordsModel":
        """Fetches a batch of records from Argilla."""
        return datasets_api_v1.get_records(
            client=self._client,
            id=self._dataset.id,
            offset=offset,
            limit=limit,
        ).parsed

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def add(
        self,
        records: Union[FeedbackRecord, Dict[str, Any], List[Union[FeedbackRecord, Dict[str, Any]]]],
        show_progress: bool = True,
    ) -> None:
        """Pushes a list of `FeedbackRecord`s to Argilla.

        Args:
            records: can be a single `FeedbackRecord`, a list of `FeedbackRecord`,
                a single dictionary, or a list of dictionaries. If a dictionary is provided,
                it will be converted to a `FeedbackRecord` internally.
            show_progress: Whether to show a `tqdm` progress bar while pushing the records.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            Exception: If the pushing of the records to Argilla fails.
        """
        records = self._dataset._parse_and_validate_records(records)
        for i in trange(
            0, len(records), PUSHING_BATCH_SIZE, desc="Pushing records to Argilla...", disable=not show_progress
        ):
            records_batch = []
            for record in records[i : i + PUSHING_BATCH_SIZE]:
                if record.suggestions:
                    for suggestion in record.suggestions:
                        suggestion.question_id = self.__question_name2id[suggestion.question_name]
                records_batch.append(
                    record.dict(exclude={"id": ..., "suggestions": {"__all__": {"question_name"}}}, exclude_none=True)
                )
            datasets_api_v1.add_records(
                client=self._client,
                id=self._dataset.id,
                records=records_batch,
            )

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete(self, records: List[RemoteFeedbackRecord]) -> None:
        """Deletes a list of `RemoteFeedbackRecord`s from Argilla.

        Args:
            records: A list of `RemoteFeedbackRecord`s to delete from Argilla.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: If the deletion of the records from Argilla fails.
        """
        num_records = len(records)
        for start in range(0, num_records, DELETE_DATASET_RECORDS_MAX_NUMBER):
            end = min(start + DELETE_DATASET_RECORDS_MAX_NUMBER, num_records)
            try:
                datasets_api_v1.delete_records(
                    client=self._client,
                    id=self._dataset.id,
                    record_ids=[record.id for record in records[start:end]],
                )
            except Exception as e:
                raise RuntimeError("Failed to remove records from Argilla") from e


class RemoteFeedbackDataset(RemoteFeedbackDatasetBase[RemoteFeedbackRecords]):
    records_cls = RemoteFeedbackRecords

    def __init__(
        self,
        *,
        client: "httpx.Client",
        id: "UUID",
        name: str,
        workspace: "Workspace",
        fields: List["AllowedFieldTypes"],
        questions: List["AllowedQuestionTypes"],
        guidelines: Optional[str] = None,
    ) -> None:
        super().__init__(
            client=client,
            id=id,
            name=name,
            workspace=workspace,
            fields=fields,
            questions=questions,
            guidelines=guidelines,
        )

    def filter_by(
        self, response_status: Union[FeedbackResponseStatusFilter, List[FeedbackResponseStatusFilter]]
    ) -> FilteredRemoteFeedbackDataset:
        """Filters the current `RemoteFeedbackDataset` based on the `response_status` of
        the responses of the records in Argilla. This method creates a new class instance
        of `FilteredRemoteFeedbackDataset` with the given filters.

        Args:
            response_status: the response status/es to filter the dataset by. Can be
                one of: draft, pending, submitted, and discarded.

        Returns:
            A new instance of `FilteredRemoteFeedbackDataset` with the given filters.
        """
        if not isinstance(response_status, list):
            response_status = [response_status]
        return FilteredRemoteFeedbackDataset(
            client=self._client,
            id=self.id,
            name=self.name,
            workspace=self.workspace,
            fields=self.fields,
            questions=self.questions,
            guidelines=self.guidelines,
            filters={
                "response_status": [status.value if hasattr(status, "value") else status for status in response_status]
            },
        )

    @allowed_for_roles(roles=[UserRole.owner, UserRole.admin])
    def delete(self) -> None:
        """Deletes the current `FeedbackDataset` from Argilla. This method is just working
        if the user has either `owner` or `admin` role.

        Raises:
            PermissionError: if the user does not have either `owner` or `admin` role.
            RuntimeError: if the `FeedbackDataset` cannot be deleted from Argilla.
        """
        try:
            datasets_api_v1.delete_dataset(client=self._client, id=self.id)
        except Exception as e:
            raise RuntimeError(f"Failed while deleting the `FeedbackDataset` from Argilla with exception: {e}") from e



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional

import typer
from pydantic import constr

from argilla.server.contexts import accounts
from argilla.server.database import AsyncSessionLocal
from argilla.server.models import User, UserRole
from argilla.server.security.model import (
    USER_PASSWORD_MIN_LENGTH,
    UserCreate,
    WorkspaceCreate,
)
from argilla.tasks import async_typer
from argilla.tasks.database.users.utils import get_or_new_workspace

USER_API_KEY_MIN_LENGTH = 8


class UserCreateForTask(UserCreate):
    api_key: Optional[constr(min_length=USER_API_KEY_MIN_LENGTH)]
    workspaces: Optional[List[WorkspaceCreate]]


def role_callback(value: str) -> str:
    try:
        return UserRole(value).value
    except ValueError:
        raise typer.BadParameter("Only Camila is allowed")


def password_callback(password: str = None) -> str:
    # if password is None:
    #     raise typer.BadParameter("Password must be specified.")
    # if len(password)<USER_PASSWORD_MIN_LENGTH:
    #     raise typer.BadParameter(f"Password must be at least {USER_PASSWORD_MIN_LENGTH} characters long.")
    return password


def api_key_callback(api_key: str) -> str:
    # if api_key and len(api_key)<USER_PASSWORD_MIN_LENGTH:
    #     raise typer.BadParameter(f"API key must a string with a minimum length of {USER_API_KEY_MIN_LENGTH} characters.")
    return api_key


async def create(
    first_name: str = typer.Option(default=None, help="First name as a string."),
    username: str = typer.Option(
        default=None,
        prompt=True,
        help="Username as a lowercase string without spaces allowing letters, numbers, dashes and underscores.",
    ),
    role: UserRole = typer.Option(
        prompt=True,
        default=UserRole.annotator.value,
        show_default=True,
        help="Role for the user.",
    ),
    password: str = typer.Option(
        default=None,
        prompt=True,
        callback=password_callback,
        help=f"Password as a string with a minimum length of {USER_PASSWORD_MIN_LENGTH} characters.",
    ),
    last_name: str = typer.Option(default=None, help="Last name as a string."),
    api_key: Optional[str] = typer.Option(
        default=None,
        callback=api_key_callback,
        help=f"API key as a string with a minimum length of {USER_API_KEY_MIN_LENGTH} characters. If not specified a secure random API key will be generated",
    ),
    workspace: List[str] = typer.Option(
        default=[], help="A workspace that the user will be a member of (can be used multiple times)."
    ),
):
    """Creates a new user in the Argilla database with provided parameters"""
    async with AsyncSessionLocal() as session:
        if await accounts.get_user_by_username(session, username):
            typer.echo(f"User with username {username!r} already exists in database. Skipping...")
            return

        if await accounts.get_user_by_api_key(session, api_key):
            typer.echo(f"User with api_key {api_key!r} already exists in database. Skipping...")
            return

        user_create = UserCreateForTask(
            first_name=first_name,
            last_name=last_name,
            username=username,
            role=role,
            password=password,
            api_key=api_key,
            workspaces=[WorkspaceCreate(name=workspace_name) for workspace_name in workspace],
        )

        user = await User.create(
            session,
            first_name=user_create.first_name,
            last_name=user_create.last_name,
            username=user_create.username,
            role=user_create.role,
            password_hash=accounts.hash_password(user_create.password),
            api_key=user_create.api_key,
            workspaces=[await get_or_new_workspace(session, workspace.name) for workspace in user_create.workspaces],
        )

        typer.echo("User successfully created:")
        typer.echo(f"• first_name: {user.first_name!r}")
        if user.last_name:
            typer.echo(f"• last_name: {user.last_name!r}")
        typer.echo(f"• username: {user.username!r}")
        typer.echo(f"• role: {user.role.value!r}")
        typer.echo(f"• api_key: {user.api_key!r}")
        typer.echo(f"• workspaces: {[workspace.name for workspace in user.workspaces]!r}")


if __name__ == "__main__":
    async_typer.run(create)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter, Body, Depends, Security

from argilla.server.apis.v0.helpers import deprecate_endpoint
from argilla.server.apis.v0.models.commons.params import DATASET_NAME_PATH_PARAM, CommonTaskHandlerDependencies
from argilla.server.apis.v0.models.dataset_settings import TextClassificationSettings
from argilla.server.apis.v0.validators.text_classification import DatasetValidator
from argilla.server.commons.models import TaskType
from argilla.server.models import User
from argilla.server.security import auth
from argilla.server.services.datasets import DatasetsService


def configure_router(router: APIRouter):
    task = TaskType.text_classification
    base_endpoint = f"/{task}/{{name}}/settings"
    new_base_endpoint = f"/{{name}}/{task}/settings"

    @deprecate_endpoint(
        path=base_endpoint,
        new_path=new_base_endpoint,
        router_method=router.get,
        name=f"get_dataset_settings_for_{task}",
        operation_id=f"get_dataset_settings_for_{task}",
        description=f"Get the {task} dataset settings",
        response_model_exclude_none=True,
        response_model=TextClassificationSettings,
    )
    async def get_dataset_settings(
        name: str = DATASET_NAME_PATH_PARAM,
        ws_params: CommonTaskHandlerDependencies = Depends(),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> TextClassificationSettings:
        found_ds = await datasets.find_by_name(
            user=current_user,
            name=name,
            workspace=ws_params.workspace,
            task=task,
        )

        settings = await datasets.get_settings(
            user=current_user, dataset=found_ds, class_type=TextClassificationSettings
        )
        return settings

    @router.patch(
        path=new_base_endpoint,
        name=f"save_dataset_settings_for_{task}",
        operation_id=f"save_dataset_settings_for_{task}",
        description=f"Save the {task} dataset settings",
        response_model_exclude_none=True,
        response_model=TextClassificationSettings,
    )
    async def save_settings(
        request: TextClassificationSettings = Body(..., description=f"The {task} dataset settings"),
        name: str = DATASET_NAME_PATH_PARAM,
        ws_params: CommonTaskHandlerDependencies = Depends(),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        validator: DatasetValidator = Depends(DatasetValidator.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> TextClassificationSettings:
        found_ds = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task,
            workspace=ws_params.workspace,
        )
        await validator.validate_dataset_settings(user=current_user, dataset=found_ds, settings=request)
        settings = await datasets.save_settings(
            user=current_user,
            dataset=found_ds,
            settings=TextClassificationSettings.parse_obj(request.dict()),
        )
        return settings

    # TODO: This will be remove in next iteration
    router.put(
        path=base_endpoint,
        name=f"old_save_dataset_settings_for_{task}",
        operation_id=f"old_save_dataset_settings_for_{task}",
        description=f"Save the {task} dataset settings",
        deprecated=True,
        response_model_exclude_none=True,
        response_model=TextClassificationSettings,
    )(save_settings)
    router.put(
        path=new_base_endpoint,
        name=f"new_save_dataset_settings_for_{task}_put",
        operation_id=f"new_save_dataset_settings_for_{task}_put",
        description=f"Save the {task} dataset settings",
        deprecated=True,
        response_model_exclude_none=True,
        response_model=TextClassificationSettings,
    )(save_settings)

    @deprecate_endpoint(
        path=base_endpoint,
        new_path=new_base_endpoint,
        router_method=router.delete,
        operation_id=f"delete_{task}_settings",
        name=f"delete_{task}_settings",
        description=f"Delete {task} dataset settings",
    )
    async def delete_settings(
        name: str = DATASET_NAME_PATH_PARAM,
        ws_params: CommonTaskHandlerDependencies = Depends(),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        user: User = Security(auth.get_current_user),
    ) -> None:
        found_ds = await datasets.find_by_name(
            user=user,
            name=name,
            task=task,
            workspace=ws_params.workspace,
        )

        await datasets.delete_settings(
            user=user,
            dataset=found_ds,
        )

    return router



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Optional

from typing_extensions import Literal

from argilla.client.feedback.training.base import ArgillaTrainerSkeleton
from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.spacy import ArgillaSpaCyTrainer as ArgillaSpaCyTrainerV1
from argilla.training.spacy import ArgillaSpaCyTransformersTrainer as ArgillaSpaCyTransformersTrainerV1
from argilla.training.spacy import _ArgillaSpaCyTrainerBase as _ArgillaSpaCyTrainerBaseV1
from argilla.utils.dependency import require_version


class _ArgillaSpaCyTrainerBase(_ArgillaSpaCyTrainerBaseV1, ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaSpaCyTrainer")
    _logger.setLevel(logging.INFO)

    require_version("spacy")

    def __init__(
        self,
        language: Optional[str] = None,
        gpu_id: Optional[int] = -1,
        model: Optional[str] = None,
        optimize: Literal["efficiency", "accuracy"] = "efficiency",
        *args,
        **kwargs,
    ) -> None:
        """Initialize the `_ArgillaSpaCyTrainerBase` class.

        Args:
            dataset: A `spacy.tokens.DocBin` object or a tuple of `spacy.tokens.DocBin` objects.
            record_class:
                A `TextClassificationRecord`, `TokenClassificationRecord`, or `Text2TextRecord`
                object. Defaults to None.
            seed: A `int` with the seed for the random number generator. Defaults to None.
            multi_label: A `bool` indicating whether the task is multi-label or not. Defaults to False.
            language:
                A `str` with the `spaCy` language code e.g. "en". See all the supported languages and their
                codes in `spaCy` at https://spacy.io/usage/models#languages. Defaults to None.
            gpu_id:
                the GPU ID to use. Defaults to -1, which means that the CPU will be used by default.
                GPU IDs start in 0, which stands for the default GPU in the system, if available.
            model:
                A `str` with the `spaCy` model name to use. If it contains vectors it
                can also be used for training/fine-tuning, e.g. "en_core_web_lg"
                contains vectors, while "en_core_web_sm" doesn't. Defaults to None.
            optimize:
                A `str` with the optimization strategy to use. Either "efficiency" or "accuracy".
                Defaults to "efficiency", which means that the model will be smaller, faster,
                and use less memory, but it will be less accurate. If "accuracy" is used, the model
                will be larger, slower, and use more memory, but it will be more accurate.
                Defaults to "efficiency".

        Raises:
            NotImplementedError: If the `record_class` is not supported or if the
                `init_training_args` method has not been implemented.
        """
        ArgillaTrainerSkeleton.__init__(self, *args, **kwargs)
        import spacy

        self._nlp = None
        self._model = model

        self.config = {}

        if self._record_class == TokenClassificationRecord:
            self._column_mapping = {
                "text": "text",
                "token": "tokens",
                "ner_tags": "ner_tags",
            }
            self._pipeline = ["ner"]
        elif self._record_class == TextClassificationRecord:
            if self._multi_label:
                self._column_mapping = {"text": "text", "binarized_label": "label"}
                self._pipeline = ["textcat_multilabel"]
            else:
                self._column_mapping = {"text": "text", "label": "label"}
                self._pipeline = ["textcat"]
        else:
            raise NotImplementedError("`rg.Text2TextRecord` is not supported yet.")

        self._train_dataset, self._eval_dataset = (
            self._dataset if isinstance(self._dataset, tuple) and len(self._dataset) > 1 else (self._dataset, None)
        )
        self._train_dataset_path = "./train.spacy"
        self._eval_dataset_path = "./dev.spacy" if self._eval_dataset else "./train.spacy"

        self.language = language or "en"
        self.optimize = optimize

        self.gpu_id = gpu_id
        self.use_gpu = False
        if self.gpu_id != -1:
            self.use_gpu = spacy.prefer_gpu(self.gpu_id)

        if self.use_gpu:
            try:
                require_version("torch")
                self.has_torch = True
            except Exception:
                self.has_torch = False

            try:
                require_version("tensorflow")
                self.has_tensorflow = True
            except Exception:
                self.has_tensorflow = False

            if not self.has_torch and not self.has_tensorflow:
                self._logger(
                    "Either `torch` or `tensorflow` need to be installed to use the"
                    " GPU, since any of those is required as the GPU allocator. Falling"
                    " back to the CPU."
                )
                self.use_gpu = False
                self.gpu_id = -1

        self.init_training_args()


class ArgillaSpaCyTrainer(ArgillaSpaCyTrainerV1, _ArgillaSpaCyTrainerBase):
    def __init__(self, freeze_tok2vec: bool = False, **kwargs) -> None:
        """Initialize the `ArgillaSpaCyTrainer` class.

        Args:
            freeze_tok2vec: A `bool` indicating whether to freeze the `tok2vec` weights
                during the training. Defaults to False.
            **kwargs: The `ArgillaSpaCyTrainerBase` arguments.

        Examples:
            >>> from argilla import ArgillaSpaCyTrainer
            >>> trainer = ArgillaSpaCyTrainer(
        """
        self.freeze_tok2vec = freeze_tok2vec
        _ArgillaSpaCyTrainerBase.__init__(self, **kwargs)


class ArgillaSpaCyTransformersTrainer(ArgillaSpaCyTransformersTrainerV1, _ArgillaSpaCyTrainerBase):
    def __init__(self, update_transformer: bool = True, **kwargs) -> None:
        """Initialize the `ArgillaSpaCyTransformersTrainer` class.

        Args:
            update_transformer: A `bool` indicating whether to update the transformer
                weights during the training. Defaults to True.
            **kwargs: The `ArgillaSpaCyTrainerBase` arguments.
        """
        self.update_transformer = update_transformer
        _ArgillaSpaCyTrainerBase.__init__(self, **kwargs)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Dataset models definition
"""
from datetime import datetime
from typing import Any, Dict, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field

from argilla._constants import ES_INDEX_REGEX_PATTERN
from argilla.server.commons.models import TaskType


class UpdateDatasetRequest(BaseModel):
    """
    Modifiable fields for dataset

    Attributes:
    -----------
    tags:
        Dataset tags used for better organize or classify information
    metadata:
        Extra information that could be interested to include
    """

    tags: Dict[str, str] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)


class _BaseDatasetRequest(UpdateDatasetRequest):
    name: str = Field(regex=ES_INDEX_REGEX_PATTERN, description="The dataset name")


class CreateDatasetRequest(_BaseDatasetRequest):
    task: TaskType = Field(description="The dataset task")
    workspace: Optional[str] = None


class CopyDatasetRequest(_BaseDatasetRequest):
    """
    Request body for copy dataset operation
    """

    target_workspace: Optional[str] = None


class Dataset(CreateDatasetRequest):
    id: Union[str, UUID]
    task: TaskType
    owner: str = Field(description="Deprecated. Use `workspace` instead. Will be removed in v1.5.0")
    workspace: str

    tags: Dict[str, str] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)
    created_at: datetime
    created_by: Optional[str] = Field(description="The argilla user that created the dataset")
    last_updated: datetime

    class Config:
        orm_mode = True



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import warnings
from typing import List, Optional

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

from pydantic import BaseModel, Field

try:
    from yaml import SafeLoader, dump, load
except ImportError:
    raise ImportError(
        "Please make sure to install `PyYAML` in order to use `DatasetConfig`. To do"
        " so you can run `pip install pyyaml`."
    )

from argilla.client.feedback.schemas.types import AllowedFieldTypes, AllowedQuestionTypes


class DatasetConfig(BaseModel):
    fields: List[AllowedFieldTypes]
    questions: List[Annotated[AllowedQuestionTypes, Field(..., discriminator="type")]]
    guidelines: Optional[str] = None

    def to_yaml(self) -> str:
        return dump(self.dict(exclude={"fields": {"__all__": {"id"}}, "questions": {"__all__": {"id"}}}))

    @classmethod
    def from_yaml(cls, yaml: str) -> "DatasetConfig":
        yaml = re.sub(r"(\n\s*|)id: !!python/object:uuid\.UUID\s+int: \d+", "", yaml)
        return cls(**load(yaml, Loader=SafeLoader))


# TODO(alvarobartt): here for backwards compatibility, remove in 1.14.0
class DeprecatedDatasetConfig(BaseModel):
    fields: List[AllowedFieldTypes]
    questions: List[AllowedQuestionTypes]
    guidelines: Optional[str] = None

    def to_json(self) -> str:
        warnings.warn(
            "`DatasetConfig` can just be dumped to YAML, so make sure that you are"
            " dumping to a YAML file instead of a JSON file. `DatasetConfig` will come"
            " in YAML format from now on, instead of JSON format.",
            DeprecationWarning,
        )
        return self.json()

    @classmethod
    def from_json(cls, json: str) -> "DeprecatedDatasetConfig":
        warnings.warn(
            "`DatasetConfig` can just be loaded from YAML, so make sure that you are"
            " loading a YAML file instead of a JSON file. `DatasetConfig` will be dumped"
            " as YAML from now on, instead of JSON.",
            DeprecationWarning,
        )
        return cls.parse_raw(json)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Tuple

from pydantic import BaseModel, Field

from argilla._constants import PROTECTED_METADATA_FIELD_PREFIX
from argilla.logging import LoggingMixin
from argilla.server.commons.models import TaskType
from argilla.server.daos.backend.base import IndexNotFoundError, InvalidSearchError
from argilla.server.daos.backend.client_adapters.base import IClientAdapter
from argilla.server.daos.backend.client_adapters.factory import ClientAdapterFactory
from argilla.server.daos.backend.mappings.datasets import DATASETS_INDEX_NAME, datasets_index_mappings
from argilla.server.daos.backend.mappings.helpers import mappings, tasks_common_mappings, tasks_common_settings
from argilla.server.daos.backend.mappings.text2text import text2text_mappings
from argilla.server.daos.backend.mappings.text_classification import text_classification_mappings
from argilla.server.daos.backend.mappings.token_classification import token_classification_mappings
from argilla.server.daos.backend.metrics import ALL_METRICS
from argilla.server.daos.backend.metrics.base import ElasticsearchMetric
from argilla.server.daos.backend.search.model import (
    BackendRecordsQuery,
    BaseDatasetsQuery,
    BaseQuery,
    BaseRecordsQuery,
    SortableField,
    SortConfig,
)
from argilla.server.errors import BadRequestError, EntityNotFoundError
from argilla.server.errors.task_errors import MetadataLimitExceededError
from argilla.server.settings import settings


def dataset_records_index(dataset_id: str) -> str:
    index_mame_template = settings.dataset_records_index_name
    return index_mame_template.format(dataset_id)


class PaginatedSortInfo(BaseModel):
    shuffle: bool = False
    sort_by: List[SortableField] = Field(default_factory=list)
    next_search_params: Optional[Any] = None


class GenericElasticEngineBackend(LoggingMixin):
    """
    Encapsulates logic about the communication, queries and index mapping
    transformations between DAOs layer and the elasticsearch backend.
    """

    _INSTANCE = None

    # TODO(@frascuchon): Once id is included as keyword in datasets index, we can discard this
    __MAX_NUMBER_OF_LISTED_DATASETS__ = 2500

    @classmethod
    def get_instance(cls) -> "GenericElasticEngineBackend":
        """
        Creates an instance of ElasticsearchBackend.

        This function is used in fastapi for resolve component dependencies.

        See <https://fastapi.tiangolo.com/tutorial/dependencies/>

        Returns
        -------

        """

        if not cls._INSTANCE:
            instance = cls(
                client=ClientAdapterFactory.get(
                    hosts=settings.elasticsearch,
                    index_shards=settings.es_records_index_shards,
                    ssl_verify=settings.elasticsearch_ssl_verify,
                    ca_path=settings.elasticsearch_ca_path,
                ),
                metrics={**ALL_METRICS},
                mappings={
                    TaskType.text_classification: text_classification_mappings(),
                    TaskType.token_classification: token_classification_mappings(),
                    TaskType.text2text: text2text_mappings(),
                },
            )
            cls._INSTANCE = instance

        return cls._INSTANCE

    def __init__(
        self,
        client: IClientAdapter,
        metrics: Dict[str, ElasticsearchMetric] = None,
        mappings: Dict[str, Dict[str, Any]] = None,
    ):
        self.__client__ = client
        self.__defined_metrics__ = metrics or {}
        self.__tasks_mappings__ = mappings

        self._common_records_mappings = tasks_common_mappings()
        self._common_records_settings = tasks_common_settings()

    @property
    def client(self) -> IClientAdapter:
        """The elasticsearch client"""
        return self.__client__

    def find_metric_by_id(self, metric_id: str) -> Optional[ElasticsearchMetric]:
        metric = self.__defined_metrics__.get(metric_id)
        if not metric:
            raise EntityNotFoundError(
                name=metric_id,
                type=ElasticsearchMetric,
            )
        return metric

    def get_task_mapping(self, task: TaskType) -> Dict[str, Any]:
        return self.__tasks_mappings__[task]

    def compute_metric(
        self,
        id: str,
        metric_id: str,
        query: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        index = dataset_records_index(id)
        return self._compute_metric(
            index=index,
            metric_id=metric_id,
            query=query,
            params=params,
        )

    def get_schema(self, id: str) -> Dict[str, Any]:
        index = dataset_records_index(id)
        return self.client.get_index_schema(index=index)

    async def update_records_content(
        self,
        id: str,
        content: Dict[str, Any],
        query: Optional[BaseDatasetsQuery],
    ) -> Tuple[int, int]:
        index = dataset_records_index(id)
        response = self.client.update_docs_by_query(
            index=index,
            data=content,
            query=query,
        )
        total, updated = response["total"], response["updated"]
        return total, updated

    async def delete_records_by_query(
        self,
        id: str,
        query: Optional[BaseDatasetsQuery],
    ) -> Tuple[int, int]:
        index = dataset_records_index(id)
        total, deleted = self.client.delete_docs_by_query(
            index=index,
            query=query,
        )
        return total, deleted

    def search_records(
        self,
        id: str,
        query: BackendRecordsQuery = BaseRecordsQuery(),
        sort: SortConfig = SortConfig(),
        record_from: int = 0,
        size: int = 100,
        exclude_fields: List[str] = None,
        enable_highlight: bool = True,
    ) -> Tuple[int, List[Dict[str, Any]]]:
        index = dataset_records_index(id)

        if not sort.sort_by and sort.shuffle is False:
            sort.sort_by = [SortableField(id="id")]  # Default sort by id

        results = self.client.search_docs(
            index=index,
            query=query,
            sort=sort,
            doc_from=record_from,
            size=size,
            exclude_fields=exclude_fields,
            enable_highlight=enable_highlight,
        )

        total = results["total"]
        records = results["docs"]

        return total, records

    def scan_records(
        self,
        id: str,
        query: BackendRecordsQuery,
        sort: PaginatedSortInfo,
        limit: Optional[int] = None,
        include_fields: Optional[List[str]] = None,
        exclude_fields: Optional[List[str]] = None,
    ) -> Iterable[Dict[str, Any]]:
        index = dataset_records_index(id)

        yield from self.client.scan_docs(
            index=index,
            query=query,
            sort=SortConfig(shuffle=sort.shuffle, sort_by=sort.sort_by),
            size=limit,
            search_from_params=sort.next_search_params,
            fetch_once=sort.shuffle,
            include_fields=include_fields,
            exclude_fields=exclude_fields,
            enable_highlight=True,
        )

    def open(self, id: str):
        self.client.open_index(dataset_records_index(id))

    def create_dataset(
        self,
        id: str,
        task: TaskType,
        metadata_values: Optional[Dict[str, Any]] = None,
        vectors_cfg: Optional[Dict[str, Any]] = None,
        force_recreate: bool = False,
    ) -> None:
        _mappings = self._common_records_mappings
        task_mappings = self.get_task_mapping(task).copy()
        for k in task_mappings:
            if isinstance(task_mappings[k], list):
                _mappings[k] = [*_mappings.get(k, []), *task_mappings[k]]
            else:
                _mappings[k] = {**_mappings.get(k, {}), **task_mappings[k]}

        index = dataset_records_index(id)
        self.client.create_index(
            index=index,
            settings=self._common_records_settings,
            mappings={**self._common_records_mappings, **_mappings},
            force_recreate=force_recreate,
        )
        if metadata_values:
            self._configure_metadata_fields(
                index=index,
                metadata_values=metadata_values,
            )

        if vectors_cfg:
            self._configure_vectors_fields(
                index=index,
                vectors_cfg=vectors_cfg,
            )

    def _configure_vectors_fields(
        self,
        index: str,
        vectors_cfg: dict,
    ):
        def _check_max_number_of_vectors():
            vectors = self.client.get_property_type(
                index=index,
                property_name="vectors",
                drop_extra_props=True,
            )
            vector_names = {key for key in vectors.keys() if "vector" in vectors[key]}
            for key in vectors_cfg.keys():
                vector_names.add(f"vectors.{key}.value")

            if len(vector_names) > settings.vectors_fields_limit:
                raise BadRequestError(
                    detail=f"Cannot create more than {settings.vectors_fields_limit} " "kind of vectors per dataset"
                )

        _check_max_number_of_vectors()

        self.client.configure_index_vectors(
            index=index,
            vectors=vectors_cfg,
        )

    def _configure_metadata_fields(self, index: str, metadata_values: Dict[str, Any]):
        def check_metadata_length(metadata_length: int = 0):
            if metadata_length > settings.metadata_fields_limit:
                raise MetadataLimitExceededError(
                    length=metadata_length,
                    limit=settings.metadata_fields_limit,
                )

        def detect_nested_type(v: Any) -> bool:
            """Returns True if value match as nested value"""
            return isinstance(v, list) and isinstance(v[0], dict)

        check_metadata_length(len(metadata_values))
        check_metadata_length(
            len(
                {
                    *self.client.get_property_type(
                        index=index,
                        property_name="metadata",
                        drop_extra_props=True,
                    ),
                    *[f"metadata.{k}" for k in metadata_values.keys()],
                }
            )
        )

        index_mappings = {}
        for field, value in metadata_values.items():
            if field.startswith(PROTECTED_METADATA_FIELD_PREFIX):
                index_mappings[f"metadata.{field}"] = mappings.protected_non_searchable_field()
            elif detect_nested_type(value):
                index_mappings[f"metadata.{field}"] = mappings.nested_field()

        self.client.set_index_mappings(index, properties=index_mappings)

    def delete(self, id: str):
        index = dataset_records_index(id)
        try:
            self.client.delete_index(
                index=index,
                raises_error=True,
            )
        except InvalidSearchError:
            # It's an alias --> DELETE from original index
            original_index = self._old_dataset_index(id)
            self.client.delete_index_alias(
                index=original_index,
                alias=index,
            )
        finally:
            # TODO: This should be move to the service layer and drop from datasets
            self.client.delete_index_document(
                index=DATASETS_INDEX_NAME,
                id=id,
            )

    def _old_dataset_index(self, id):
        return settings.old_dataset_records_index_name.format(id)

    def copy(self, id_from: str, id_to: str):
        index_from = dataset_records_index(id_from)
        index_to = dataset_records_index(id_to)

        self.client.copy_index(
            source_index=index_from,
            target_index=index_to,
        )

    def close(self, id: str):
        return self.client.close_index(index=dataset_records_index(id))

    def create_datasets_index(self, force_recreate: bool = False):
        self.client.create_index(
            index=DATASETS_INDEX_NAME,
            force_recreate=force_recreate,
            mappings=datasets_index_mappings(),
        )
        # TODO: Remove this section of code
        if settings.enable_migration:
            try:
                self._migrate_from_rubrix()
            except IndexNotFoundError:
                pass  # Nothing to migrate

    def _migrate_from_rubrix(self):
        source_index = settings.old_dataset_index_name
        target_index = DATASETS_INDEX_NAME

        try:
            self.client.copy_index(
                source_index=source_index,
                target_index=target_index,
                reindex=True,
            )
            for doc in self.client.scan_docs(index=source_index, query=BaseQuery(), sort=SortConfig()):
                dataset_id = doc["id"]
                index = self._old_dataset_index(dataset_id)
                alias = dataset_records_index(dataset_id=dataset_id)
                self._update_dynamic_mapping(index)
                self.client.create_index_alias(
                    index=index,
                    alias=alias,
                )
        except IndexNotFoundError:
            pass  # Nothing to migrate

    def _update_dynamic_mapping(self, index: str):
        self.client.set_index_mappings(
            index=index,
            dynamic=False,
        )

    def list_datasets(self, query: BaseDatasetsQuery):
        return self.client.scan_docs(
            index=DATASETS_INDEX_NAME,
            query=query,
            sort=SortConfig(),
            fetch_once=True,
            size=self.__MAX_NUMBER_OF_LISTED_DATASETS__,
        )

    def find_record_by_id(self, dataset_id: str, record_id: str):
        index = dataset_records_index(dataset_id)
        return self.client.get_index_document_by_id(
            index=index,
            id=record_id,
        )

    def add_dataset_document(
        self,
        id: str,
        document: Dict[str, Any],
    ):
        self.client.upsert_index_document(
            index=DATASETS_INDEX_NAME,
            id=id,
            document=document,
        )

    def update_dataset_document(
        self,
        id: str,
        document: Dict[str, Any],
    ):
        self.client.upsert_index_document(
            index=DATASETS_INDEX_NAME,
            id=id,
            document=document,
            partial_update=True,
        )

    def update_record(
        self,
        dataset_id: str,
        record_id: str,
        content: dict,
    ):
        index = dataset_records_index(dataset_id)
        self.client.upsert_index_document(
            index=index,
            id=record_id,
            document=content,
        )

    def find_dataset(
        self,
        id: str,
    ):
        document = self.client.get_index_document_by_id(
            index=DATASETS_INDEX_NAME,
            id=id,
        )
        return document

    def _compute_metric(
        self,
        index: str,
        metric_id: str,
        query: Optional[BaseQuery] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict:
        metric = self.find_metric_by_id(metric_id)
        data = self.client.compute_index_metric(
            index=index,
            metric=metric,
            query=query,
            params=params or {},
        )
        return data

    def remove_dataset_field(self, id: str, field: str):
        self.client.drop_document_property(
            index=DATASETS_INDEX_NAME,
            id=id,
            property=field,
        )

    def add_dataset_records(self, id: str, documents: List[dict]) -> int:
        index = dataset_records_index(id)

        return self.client.index_documents(index=index, docs=documents)
